import customtkinter
import tkinter
import tkinter.font
import tkinter.ttk
import os, sys, platform
import signal
import screeninfo
from .assets import Assets
from .shared import SHARED
from .widgets.labels import *
import pandas as pd
from .tabs.trashdetectiontab import TRASHTAB
from .tabs.gausian_process_tab_nomap import GAUSIANSENSORTAB
from .tabs.mission_tab import MISSIONTAB
from .tabs.dijkstratab import DIJKSTRATAB
from .tabs.map_generation_tab import MAP_GENERATORTAB
# from tabs.map_generation_tab import MAP_GENERATORTAB
# import machine this is for hard reset (machine.reset())
version = "v1.0.0"
max_tab_idx = -1
def next_tab_idx():
    global max_tab_idx
    max_tab_idx += 1
    return max_tab_idx
class ACETI_GUI(tkinter.Tk):
    def __init__(self):
        self.platform=platform.system()
            #get platform
        super().__init__()
        #INIT SHAPE  todo:use customTK
        # customtkinter.set_appearance_mode("System")  # Modes: system (default), light, dark
        # customtkinter.set_default_color_theme("blue")  # Themes: blue (default), dark-blue, green
        self.title("ACETI GUI")
        # schedule a rescale after init and keep handle so we can cancel on exit
        self._rescale_handle = self.after(1000, self.rescale) #after init, scale window
        # elif platform.system() == “Windows”:
        self.configure(bg = 'white')
        self.option_add('*tearOff', False)  # Deshabilita submenús flotantes
        self.max_tab_idx = 0

        #VARIABLES FOR FUNCTIONS
        self.fullscreenstate = False
        # create Assets bound to this Tk instance so PhotoImage objects have a master
        self.assets = Assets(master=self)
        self.shared = SHARED()
        # self.toggle_fullscreen()
        # self.toggle_fullscreen()
        #get screen size
        monitors = screeninfo.get_monitors()
        self.screenheight=monitors[0].height
        self.screenwidth=monitors[0].width
        self.screenyoffset=monitors[0].y
        self.screenxoffset=monitors[0].x
        for m in reversed(monitors):
            if m.x <= self.winfo_x() <= m.width + m.x and m.y <= self.winfo_y() <= m.height + m.y:
                self.screenheight=m.height
                self.screenwidth=m.width
                self.screenyoffset=m.y
                self.screenxoffset=m.x

        # Style for white ttk PanedWindow
        self.option_add('*Background', 'white')
        self.option_add('*Foreground', 'black')
        self.option_add('*Button.Background', 'white')
        self.option_add('*Button.Foreground', 'black')
        self.option_add('*Label.Background', 'white')
        self.option_add('*Menu.background', 'white')
        self.option_add('*Menu.foreground', 'black')
        self.option_add('*Menu.activeBackground', '#e6e6e6')

        style = tkinter.ttk.Style()
        style.configure('.', background='white', foreground='black')
        style.configure('TFrame', background='white')
        style.configure('TPanedwindow', background='white')
        style.configure('White.TPanedwindow', background='white')
        style.configure('TLabel', background='white', foreground='black')
        style.configure('TButton', background='white', foreground='black')
        style.configure('TNotebook', background='white')
        style.configure('TNotebook.Tab', background='white')
        style.configure('TCombobox', fieldbackground='white', background='white', foreground='black')
        style.map('TCombobox',
                  fieldbackground=[('readonly', 'white')],
                  background=[('active', 'white'), ('!disabled', 'white')])

        self.mainFrame = tkinter.ttk.PanedWindow(orient="vertical", style='White.TPanedwindow')
        self.mainFrame.pack(side='top', fill='both', expand=True,)
        self.iconphoto(self, self.assets.aceti_icon)

        #menubar
        self.toolbar = tkinter.Frame(self.mainFrame, relief='raised', bd=2, bg="#E5E5E5")
        tkinter.Label(self.toolbar, text=version).pack(side="right", fill='y', padx=0, pady=2)
        
        self.mainFrame.add(self.toolbar)
        self.menubar_buttons=[]

        ####################################################
        ############## different tabs
        ####################################################
        self.tabs=[]
        
        #create trash map tab
        self.button_trash_detection = tkinter.Button(self.toolbar, image=self.assets.icon_water_trash, command=lambda a=next_tab_idx(): self.change_tab(a), width=40, height=40)
        self.button_trash_detection.pack(side='left', expand=False,)
        self.menubar_buttons.append(self.button_trash_detection)   
        self.TRASHTAB = TRASHTAB(parent=self) 
        self.tabs.append(self.TRASHTAB)

        #create gaussian sensor tab
        self.button_gaussian_sensor = tkinter.Button(self.toolbar, image=self.assets.icon_gaussian_process, command=lambda a=next_tab_idx(): self.change_tab(a), width=40, height=40)
        self.menubar_buttons.append(self.button_gaussian_sensor)
        self.button_gaussian_sensor.pack(side='left', expand=False,)
        self.GAUSIANSENSORTAB=GAUSIANSENSORTAB(parent=self)
        self.tabs.append(self.GAUSIANSENSORTAB)

        # create mission tab
        self.button_mission = tkinter.Button(self.toolbar, image=self.assets.icon_path_planning, command=lambda a=next_tab_idx(): self.change_tab(a), width=40, height=40)
        self.menubar_buttons.append(self.button_mission)
        self.button_mission.pack(side='left', expand=False,)
        self.MISSIONTAB = MAP_GENERATORTAB(parent=self)
        self.tabs.append(self.MISSIONTAB)

        self.button_dijkstra = tkinter.Button(self.toolbar, image=self.assets.icon_path_planning, command=lambda a=next_tab_idx(): self.change_tab(a), width=40, height=40)
        self.menubar_buttons.append(self.button_dijkstra)
        self.button_dijkstra.pack(side='left', expand=False,)
        self.DIJKSTRATAB = DIJKSTRATAB(parent=self)
        self.tabs.append(self.DIJKSTRATAB)

        #generate map tab
        # self.map_generation_but = tkinter.Button(self.toolbar, image=self.assets.icon_map_gen, command=lambda a=next_tab_idx(): self.change_tab(a), width=40, height=40)
        # self.menubar_buttons.append(self.map_generation_but)
        # self.map_generation_but.pack(side='left', expand=False,)
        # self.mapgentab = MAP_GENERATORTAB(parent=self)
        # self.tabs.append(self.mapgentab)
        
        #select init tab
        self.change_tab(3)

        #MENU BAR
        menubar = tkinter.Menu(self)
        self['menu']=menubar
        
        filemenu = tkinter.Menu(menubar)
        settings = tkinter.Menu(menubar)
        helpmenu = tkinter.Menu(menubar)
        menubar.add_cascade(menu=filemenu, label='File')
        menubar.add_cascade(menu=settings, label='Settings')
        menubar.add_cascade(menu=helpmenu, label='Help')

        #FILEMENU
        filemenu.add_separator()  # Agrega un separador
        filemenu.add_command(label='Exit', command=self.destroy, 
                            underline=0, accelerator="Ctrl+q",
                            image=self.assets.icon_trash_log, compound='left')

        #KEY BINDS
        self.bind("<F11>", self.toggle_fullscreen)
        self.bind("<Escape>", self.end_program)
        # ensure we do a clean shutdown when window is closed by the window manager
        self.protocol("WM_DELETE_WINDOW", self.end_program)
        
        #extravars
        boldfont = tkinter.font.Font(weight='bold')



    def toggle_fullscreen(self, event=None):
        self.fullscreenstate = not self.fullscreenstate  # Just toggling the boolean
        self.attributes("-fullscreen", self.fullscreenstate)
        return "break"

    def end_program(self, event=None):
        # attempt graceful shutdown: cancel scheduled 'after' callbacks from tabs and self
        try:
            # prevent re-entrance
            if getattr(self, '_closing', False):
                return "break"
            self._closing = True

            # cancel own rescale callback
            try:
                if hasattr(self, '_rescale_handle') and self._rescale_handle is not None:
                    self.after_cancel(self._rescale_handle)
            except Exception:
                pass

            # cancel known callbacks stored in tabs
            for t in getattr(self, 'tabs', []):
                # common attribute names that hold after handles
                for attr in ('play_callback', 'download_loop_object', 'gp_loop_object', 'after_object', 'mission_loop_object'):
                    try:
                        handle = getattr(t, attr, None)
                        if handle:
                            try:
                                self.after_cancel(handle)
                            except Exception:
                                pass
                            try:
                                setattr(t, attr, None)
                            except Exception:
                                pass
                    except Exception:
                        pass
                # call tab-specific close() if exists
                try:
                    if hasattr(t, 'close'):
                        t.close()
                except Exception:
                    pass

            # final quit/destroy
            try:
                self.quit()
            except Exception:
                pass
            try:
                self.destroy()
            except Exception:
                pass
            # try to explicitly delete image assets and force GC to avoid PIL PhotoImage destructor warnings
            try:
                import gc
                try:
                    if hasattr(self, 'assets'):
                        # remove references to PhotoImage objects
                        for k,v in list(self.assets.__dict__.items()):
                            try:
                                setattr(self.assets, k, None)
                            except Exception:
                                pass
                except Exception:
                    pass
                gc.collect()
            except Exception:
                pass
        except Exception:
            pass
        return "break"

    def dummy(self):
        pass

    def rescale(self, event=None):
        if self.platform == "Linux":
            self.attributes('-zoomed', True)
        else:
            self.state('zoomed')

    def change_tab(self, tab):
        for i in range(len(self.menubar_buttons)):
            if i == tab:
                self.menubar_buttons[tab].config(relief="sunken")
            else:
                self.menubar_buttons[i].config(relief="raised")
        for i in range(len(self.tabs)):
            try:
                if tab==i:
                    self.mainFrame.add(self.tabs[i])
                    self.tabs[i].open()
                else:
                    self.mainFrame.forget(self.tabs[i])
                    self.tabs[i].close()
            except:
                # error= traceback.format_exc()
                # print(f"There was an error oculting tab\n{error}")
                pass

def main(args=None):
    web=ACETI_GUI()
    print(f"Init Version {version}")
    web.mainloop() 
    print("Finishing Application")
    print("Application Finished")
    

if __name__ == '__main__':
    main()



import tkinter
import tkinter.font
import tkinter.ttk
from tkintermapview import TkinterMapView
from PIL import Image, ImageTk
import random
from tkintermapview.canvas_position_marker import CanvasPositionMarker
from tkintermapview.canvas_path import CanvasPath
import matplotlib.pyplot as plt
import matplotlib as mpl


class SHIP(CanvasPositionMarker):
    def __init__(self, name, deg_x: float, deg_y: float,text: str = None, rotation=None, parent=None, *args, **kwargs):
        
        if parent is None:
            super().__init__(self,(deg_x,deg_y),text=text,*args, **kwargs)
            return #stop doing anything


        #store parent
        self.parent=parent
        self.map_widget=parent.map_widget

        #store bag name
        self.name=name
        
        #restore context
        super().__init__(self.map_widget,(deg_x,deg_y), text=text, image_zoom_visibility=(0, float("inf")),  icon=ImageTk.PhotoImage(self.parent.assets.image_ship),command=self.leftclick)
        
        if rotation is not None:
            self.update_rotation(rotation)
        #draw it
        self.draw()
        self.map_widget.canvas_marker_list.append(self)

    def mouse_enter(self,event=None):
        self.set_text(self.name)
        super().mouse_enter(event)
        self.draw()

    def mouse_leave(self, event=None):
        self.set_text("")
        super().mouse_leave(event)

    def leftclick(self, event=None):
        pass

    def update_rotation(self,degrees):
        self.change_icon(ImageTk.PhotoImage(self.parent.assets.image_ship.rotate(degrees)))

class TRASH(CanvasPositionMarker):
    def __init__(self, name, deg_x: float, deg_y: float,text: str = None, parent=None):
        
        if parent is None:
            super().__init__(self,(deg_x,deg_y),text=text,**kwargs)
            return #stop doing anything

        self.max_retain_data=40

        #store parent
        self.parent=parent
        self.map_widget=parent.map_widget
        self.last_value=-200
        #store bag name
        self.name=name
        
        #restore context
        super().__init__(self.map_widget,(deg_x,deg_y), text=text, image_zoom_visibility=(0, float("inf")),command=self.leftclick, icon=self.parent.assets.icon_waste[random.randint(0, self.parent.assets.waste_len-1)])
        
        #draw it
        self.draw()
        self.map_widget.canvas_marker_list.append(self)

    def mouse_enter(self,event=None):
        self.set_text(self.name)
        super().mouse_enter(event)
        self.draw()

    def mouse_leave(self, event=None):
        self.set_text("")
        super().mouse_leave(event)

    def leftclick(self, event=None):
        pass

    def update(self, value):
        if abs(value-self.last_value)>self.max_retain_data:
            self.hide_image(True)
    
    def update_position(self, lat, lon):
        self.hide_image(False)
        self.set_position(lat,lon)



class MEASURE(CanvasPositionMarker):
    def __init__(self, name, deg_x: float, deg_y: float,text: str = None, parent=None, range_=1, min_=0, **kwargs):
        
        if parent is None:
            super().__init__(self,(deg_x,deg_y),text=text,**kwargs)
            return #stop doing anything


        #store parent
        self.parent=parent
        self.map_widget=parent.map_widget
        #store bag name
        self.name=name
        
        #restore context
        # print(f"drawing a measure {((name-min_)*10)//range_}")
        super().__init__(self.map_widget,(deg_x,deg_y), text=text,icon=self.parent.assets.icon_measure[int(((name-min_)*10)//range_)], image_zoom_visibility=(0, float("inf")),command=self.leftclick, text_color="black")
        
        #draw it
        self.draw()
        self.map_widget.canvas_marker_list.append(self)
        if name == 0:
            self.hide_image(True)

    def leftclick(self, event=None):
        try:
            super().leftclick(event)
        except:
            pass

class PATH():
    def __init__(self, df, parent=None, sensor = "ph", **kwargs):
        self.colorlist=["#983604", "#CE3202","#E53602","#FE6501","#FDFC0A","#98CC05","#037F03","#3663FE","#0301CE","#300397","#010263","#030153","#020330","#000000"]
        cmap = mpl.colors.LinearSegmentedColormap.from_list("dummy",self.colorlist)
        self.df = df 
        self.parent=parent
        self.map_widget=parent.map_widget
        self.range=[0,15]
        self.shown_values=0
        self.sensor=sensor
        
        norm = mpl.colors.Normalize(vmin=self.range[0], vmax=self.range[1])
        self.m=mpl.cm.ScalarMappable(norm=norm, cmap=cmap)

        print("we got path")
        self.last_segment=None
        

    def show_up_to(self, value):
        if value == 0:
            value = self.sel_df.index.to_list()[0]
        if value not in self.sel_df.index:
            return
        int_value = self.sel_df .index.to_list().index(value)
        if int_value > self.shown_values:
            for i in self.sel_df.index.to_list()[self.shown_values:int_value]:
                segment=[self.sel_df.loc[i,"Latitude"], self.sel_df.loc[i,"Longitude"]]
                print(segment)
                if self.last_segment is not None:
                    path=CanvasPath(map_widget=self.map_widget,position_list= [segment,self.last_segment] , color=self.get_color(self.sel_df.loc[i,"Data"]), width=4)
                    path.draw()
                    self.map_widget.canvas_path_list.append(path)
                self.last_segment=segment
            self.shown_values=int_value
        else:
            print("resetted list")
            self.map_widget.delete_all_path()
            self.last_segment=None
            self.shown_values=0

        
    def select_sensor(self, sensor:str):
        self.map_widget.delete_all_path()
        self.last_segment=None
        self.shown_values=0
        cmap = mpl.colors.LinearSegmentedColormap.from_list("dummy",self.colorlist)
        self.sel_df=self.df[self.df["Sensor"] == sensor]
        if len(self.sel_df) == 0:
            print("No data for this sensor")
            return
        # Limpiar outliers
        Q1 = self.sel_df['Data'].quantile(0.25)
        Q3 = self.sel_df['Data'].quantile(0.75)
        IQR = Q3 - Q1
        self.sel_df = self.sel_df[~((self.sel_df['Data'] < (Q1 - 3 * IQR)) | (self.sel_df['Data'] > (Q3 + 3 * IQR)))]

        if "ph" in sensor.lower():
            self.range=[0,15]
        else:
            self.range=[min(self.sel_df["Data"]), max(self.sel_df["Data"])]
        norm = mpl.colors.Normalize(vmin=self.range[0], vmax=self.range[1])
        self.m=mpl.cm.ScalarMappable(norm=norm, cmap=cmap)
        self.get_color_map()


    def get_color(self, value):
        value=self.m.to_rgba(value, bytes=True)
        return "#" + '{:02X}'.format(value[0]) + '{:02X}'.format(value[1]) + '{:02X}'.format(value[2])
    
    def get_color_map(self):
        self.parent.colorbar_fig.colorbar(self.m, cax=self.parent.colorbar_ax, orientation='vertical', label='Some Units')
        self.parent.colorbar.draw()



import numpy as np
from scipy.ndimage import binary_dilation
import matplotlib.pyplot as plt

# NODOS
nodos_mask = np.load('Environment/maps/AlamilloAccess11x15mask.npy')
nodos_lat_lon = np.load('Environment/maps/AlamilloAccess11x15latlon.npy')

# "REALES"
real_mask = np.genfromtxt('Environment/maps/rawfiles/Alamillo95x216plantilla.csv', dtype=int, delimiter=' ')
raw_grid_csv = np.genfromtxt('Environment/maps/rawfiles/Alamillo95x216grid.csv', delimiter=';', dtype=str)
real_lat_long = np.zeros((*real_mask.shape, 2))
# Transform the map to a float for long and lat
for i in range(real_mask.shape[0]):
	for j in range(real_mask.shape[1]):
		real_lat_long[i, j, 0] = float(raw_grid_csv[i, j].split(',')[0])
		real_lat_long[i, j, 1] = float(raw_grid_csv[i, j].split(',')[1])

latitudes = real_lat_long[:,:, 0]
longitudes = real_lat_long[:,:, 1]

navigables_latitudes = latitudes[real_mask == 1]
navigables_longitudes = longitudes[real_mask == 1]

# Comprobar si los puntos de navegación en el mapa real están en el mapa de nodos
# Para cada punto navegable (1) en base_matrix, obtener su latitud y longitud. Y comprobar si todos son alcanzables.
cont_unreach = 0
cont_reach = 0
non_reachable_points = []
reachable_points = []
for i in range(nodos_mask.shape[0]):
	for j in range(nodos_mask.shape[1]):
		if nodos_mask[i, j] == 1:
			# Comprobar si está como navegable en el mapa de navegación
			if nodos_lat_lon[i, j, 0] in navigables_latitudes and nodos_lat_lon[i, j, 1] in navigables_longitudes:
				# Obtener los índices de la latitud y longitud en el mapa de navegación
				lat_indexs = np.where(navigables_latitudes == nodos_lat_lon[i, j, 0])[0]
				lon_indexs = np.where(navigables_longitudes == nodos_lat_lon[i, j, 1])[0]

				# Comprobar si hay algún índice en común
				if len(set(lat_indexs).intersection(set(lon_indexs))) == 0:
					print(f'No alcanzable (no común): {nodos_lat_lon[i, j, 0]}, {nodos_lat_lon[i, j, 1]}')
					cont_unreach += 1
					non_reachable_points.append((i, j))
				else:
					cont_reach += 1
					reachable_points.append((i, j))
			else:
				print(f'No alcanzable: {nodos_lat_lon[i, j, 0]}, {nodos_lat_lon[i, j, 1]}')
				cont_unreach += 1
				non_reachable_points.append((i, j))


if cont_unreach == 0:
	print(f'Todos los puntos son alcanzables: {cont_reach}')
else:
	print(f'{cont_unreach} puntos no son alcanzables de un total de {np.sum(nodos_mask)}')

# Pintar los puntos no alcanzables
plt.imshow(nodos_mask)
for point in non_reachable_points:
	plt.scatter(point[1], point[0], c='r', s=10)
for point in reachable_points:
	plt.scatter(point[1], point[0], c='g', s=10)
plt.show()


import tkinter
import tkinter.font
import tkinter.ttk
import os, sys
from tkintermapview import TkinterMapView
from datetime import datetime, date
import gc
import traceback
import numpy as np
import json
sys.path.append('../')
from ..shared import *
from ..assets import *
from ..submodules.database import Database
from ..submodules.MQTT import *
from ..widgets.labels import *
from ..widgets.highlable_calendar import *
from ..widgets.map_markers import *
from ..widgets.checkbox import *



class MISSIONTAB(tkinter.ttk.PanedWindow):
    def __init__(self, parent=None):
        if parent is None:
            parent=tkinter.Tk()
        self.parent=parent
        self.trash_markers=[]
        #inherit
        try:
            self.assets=self.parent.assets
            self.shared=self.parent.shared
        except:
            self.assets=Assets()
            self.shared=SHARED()
        
        self.database= Database()
        self.busy=False
        self.mission_loop_object = None
        self.date=date.today()
        self.on_status=True
        self.taget_sensor=self.database.sensors[0]
        self.now_status=True
        self.refreshed = False
        self.waiting=False
        
        super().__init__(orient="horizontal")

        self.mission_tab()
        # self.playback_data()

        self.ships=[SHIP("vehicle_1", 0.0, 0.0, parent=self)]
        self.last_message=json.loads('{"Latitude" : 0.0 , "Longitude" : 0.0 , "Heading" : 0.0}')
        topics=["#"]

        self.mqttConnection = MQTT( on_message=self.on_message, 
                                    topics2suscribe=topics
                                    )

    def mission_tab(self):
        self.GPS_panel = tkinter.ttk.PanedWindow(orient="vertical", width=int(self.parent.screenwidth*0.8))
        self.add(self.GPS_panel)

        
        #create map
        self.map_widget = TkinterMapView(corner_radius=0, height=int(self.parent.screenheight*0.75)+1, database_path=self.assets.map_tiles) #this widget has no automatic size
        self.map_widget.set_tile_server("https://mt0.google.com/vt/lyrs=s&hl=en&x={x}&y={y}&z={z}&s=Ga", max_zoom=22)
        self.GPS_panel.add(self.map_widget)
        self.topleft=[37.420088,-6.001346]
        self.bottomright=[37.418399,-5.997504]
        self.map_widget.fit_bounding_box(self.topleft, self.bottomright)


        self.measures=[]
        self.ships=[]

        ##create info bar
        gps_data= tkinter.ttk.PanedWindow(orient="horizontal",height=1)
        self.GPS_panel.add(gps_data)
        #last value frame
        self.sensordata = tkinter.Frame(gps_data)
        Sensors=["Turbidity", "PH", "Battery", "Temperature", "Conductivity", "Sonar"]
        aux=create_labelh_with_units(self.sensordata, "Turbidity", self.shared.Turbidity, "NTU", expand="true")
        aux=create_labelh(self.sensordata, "PH", self.shared.PH, expand="true")
        aux=create_labelh_with_units(self.sensordata, "Battery:", self.shared.Battery, "%", expand="true")
        aux=create_labelh_with_units(self.sensordata, "Temperature:", self.shared.Temperature, "ºC", expand="true")
        aux=create_labelh_with_units(self.sensordata, "Conductivity:", self.shared.Conductivity, "S/cm", expand="true")
        aux=create_labelh_with_units(self.sensordata, "Sonar:", self.shared.Sonar, "m", expand="true")
        aux=create_labelh(self.sensordata, "Date:", self.shared.Date, expand="true")
        self.sensordata.pack(side="left",expand="true",fill="both")
        self.after_object=None
        
        #play and center buttons
        buttonsframe=tkinter.Frame(gps_data, height=1, borderwidth=1) #for date and play/center
        self.center_buttom = tkinter.Button(buttonsframe, text="Center", command=self.center_map, width=20, height=1,font=tkinter.font.Font(weight='bold', size=18))
        self.center_buttom.pack(side="left",expand="false",fill="y")
        buttonsframe.pack(side="right",expand="false",fill="y")

        #map selection
        self.map_selection=tkinter.Frame(self, width=20)
        self.map_selection.pack_propagate(False)
        self.add(self.map_selection)

        self.checkboxes=[]
        for i in range(len(self.database.sensors)): #for as many sensors
            aux= Checkbox(self.map_selection, text=self.database.sensors[i], width=20, command=lambda a=i: self.draw_map(a))
            aux.pack(side="top")
            self.checkboxes.append(aux)
        self.checkboxes[0].check()

        self.playback_frame=tkinter.Frame(self.map_selection)
        self.playback_button = tkinter.Button(self.playback_frame, image = self.assets.icon_off, bd = 0, command = self.playback_data)
        self.playback_label=tkinter.Label(self.playback_frame, text="On ")
        self.playback_label.pack(side="left", anchor="e")
        self.playback_button.pack(side="left", pady = 50, anchor="w")
        self.playback_frame.pack(side="top", fill="x")

        #calendar
        self.playback_selector_frame=tkinter.Frame(self.map_selection)
        self.cal = highlable_calendar(self.playback_selector_frame, selectmode = 'day', parent=self)
        self.cal.selection_set(self.date)
        # self.parent.bind("<<UpdatePlotData>>", self.download_database)
        self.cal.pack(pady = 20, expand="false",fill="none")
        self.refresh_buttom = tkinter.Button(self.playback_selector_frame, text="refresh", command=self.refresh_database, width=8, height=1,font=tkinter.font.Font(weight='bold', size=18))
        self.refresh_buttom.pack(side="top",expand="false",fill="x")


        self.playback_selector_frame.pack()
        
        self.data_points_label=tkinter.IntVar()
        a= tkinter.Frame(self.map_selection, padx=10, pady=2)
        a.pack(side='top', fill='both', expand=True,)
        a.pack_propagate(False)
        tkinter.Label(a, text="Number of Points", font=tkinter.font.Font(weight='bold')).pack(side="left", fill='both', padx=0, pady=2)
        tkinter.Label(a, textvariable=self.data_points_label).pack(side="left", fill='both', padx=0, pady=2)

        self.status_label=tkinter.StringVar()
        a= tkinter.Frame(self.map_selection, padx=10, pady=2)
        a.pack(side='top', fill='both', expand=True,)
        a.pack_propagate(False)
        tkinter.Label(a, text="Status", font=tkinter.font.Font(weight='bold')).pack(side="left", fill='both', padx=0, pady=2)
        tkinter.Label(a, textvariable=self.status_label).pack(side="left", fill='both', padx=0, pady=2)
        self.status_label.set("Off")


    def center_map(self,event=None):
        self.map_widget.fit_bounding_box(self.topleft, self.bottomright)


    def playback_data(self, event=None):
        self.on_status = not self.on_status
        #TODO: change to data from specific date or now
        if not self.on_status:
            self.after_object = self.parent.after(500, self.refresh_particles)
            self.playback_button.config(image=self.assets.icon_on)
        else:
            if self.after_object != None:
                self.parent.after_cancel(self.after_object)
            self.after_object=None
            self.playback_button.config(image=self.assets.icon_off)
            self.status_label.set("Off")
    
    def refresh_database(self, event=None):
        self.database.refresh_dates(table="PATHS")
        self.cal.unhighlight()
        for _date in self.database.date_df:
            dat=date.fromisoformat(_date)
            self.cal.add_to_data(dat)
        self.refreshed = True

    def refresh_particles(self):
        if self.on_status ==True:
            return
        if not self.database.busy and self.waiting==False:
            self.status_label.set("Querying")
            date_=self.cal.selection_get().strftime("%Y-%m-%d")
            # query=f"SELECT * FROM wqp.OBJECT_POSITION where date(Date) = '{date_}';"
            query=f"SELECT Latitude,Longitude,Data FROM wqp.WQP where date(Date) = '{date_}' AND Latitude<>0 AND Sensor='{self.taget_sensor}';"
            self.database.query(query=query) #self.date.strftime("2024-02-27")
            self.waiting=True
        if self.database.busy:
            self.after_object = self.parent.after(1000, self.refresh_particles)
            return
        self.waiting=False
        if(len(self.database.df) == 0):
            self.status_label.set("No Data")
            self.data_points_label.set(0)

            self.after_object = self.parent.after(8000, self.refresh_particles)
            return
        self.status_label.set("Printing")

        data = self.database.df[self.database.df['Data'] != 0]
        self.data_points_label.set(len(data))
        data["Latitude"] = np.ceil((data["Latitude"].astype(float).values* 10000))/10000 -0.00005
        data["Longitude"] = np.ceil((data["Longitude"].astype(float).values* 10000))/10000 -0.00005

        if True: #calculate mean
            positions = data.groupby(['Latitude', 'Longitude']).mean()
            print(f"there are {len(positions)} different positions")

            self.map_widget.delete_all_marker()
            del self.ships
            del self.measures
            gc.collect()
            self.ships=[]
            self.measures=[]
            expected_min=positions["Data"].min()
            expected_range=positions["Data"].max()-expected_min +0.001 #avoid division by 0
            print(f"we are expecting an measure between {positions['Data'].max()} and {positions['Data'].min()}")
            for row in positions.index.values:
                self.measures.append(MEASURE(float(positions["Data"].loc[row]), float(row[0]), float(row[1]), parent=self, range_=expected_range, min_=expected_min))
            self.after_object = self.parent.after(8000, self.refresh_particles)
        else:
            positions = data[['Latitude', 'Longitude']].drop_duplicates()
            self.map_widget.delete_all_marker()
            self.ships=[]
            self.measures=[]
            expected_min=positions["Data"].min()
            expected_range=positions["Data"].max()-expected_min +0.001 #avoid division by 0
            print(f"we are expecting an measure between {positions['Data'].max()} and {positions['Data'].min()}")
            for row in positions.index.values:
                self.measures.append(MEASURE(float(data["Data"].loc[row]), float(data["Latitude"].loc[row]), float(data["Longitude"].loc[row]), parent=self, range_=expected_range, min_=expected_min))
            self.after_object = self.parent.after(15000, self.refresh_particles)
        self.status_label.set(f"Iddle {self.taget_sensor}")

    def close(self):
        # ensure playback loop is stopped and any after callback cancelled
        try:
            if self.on_status==False:
                self.playback_data()
        except Exception:
            pass
        try:
            if hasattr(self, 'after_object') and self.after_object:
                self.parent.after_cancel(self.after_object)
                self.after_object = None
        except Exception:
            pass
        # attempt to destroy map widget and free resources
        try:
            if hasattr(self, 'map_widget') and self.map_widget is not None:
                try:
                    self.map_widget.delete_all_marker()
                except Exception:
                    pass
                try:
                    self.map_widget.destroy()
                except Exception:
                    pass
                self.map_widget = None
        except Exception:
            pass

    def open(self):
        if not self.refreshed:
            self.refresh_database()

    def on_message(self, _client, user_data, msg):
        try:
            if "asv_state" in msg.topic:
                self.last_message=json.loads(msg.payload)
                if len(self.ships)>0:
                    self.ships[0].update_rotation(-self.last_message['Heading'])
                    self.ships[0].set_position(self.last_message["Latitude"], self.last_message["Longitude"])
                else:
                    self.ships.append(SHIP("vehicle_1", self.last_message["Latitude"], self.last_message["Longitude"], heading=-self.last_message['Heading'], parent=self))
            elif "database/wqp" in msg.topic:
                pass
        except:
            error= traceback.format_exc()
            print(f"There was an error parsing mqtt \n{error}")

    def draw_map(self, event=None):
        for i in self.checkboxes:
            i.uncheck()
        self.checkboxes[event].check()
        self.taget_sensor=self.database.sensors[event]
"""
    'conductivity': msg.conductivity,
    'temperature_ct': msg.temperature_ct,
    'turbidity': msg.turbidity,
    'ph': msg.ph,
    'vbat': msg.vbat,
    'lat': msg.lat,
    'lon': msg.lon,
    'date': msg.date,
    'veh_num': self.vehicle_id
"""

#for measures
"""
37.4185, -6.001
"""


# Author: David Duran Perez
# Date: May 26, 2017

# Necessary imports
from tkinter import *
from tkinter import ttk

# Colors
colorActiveTab="#CCCCCC" # Color of the active tab
colorNoActiveTab="#EBEBEB" # Color of the no active tab
# Fonts
fontLabels='Calibri'
sizeLabels2=13

class ListboxEditable(object):
    """A class that emulates a listbox, but you can also edit a field"""
    # Constructor
    def __init__(self,frameMaster,lista):
        # *** Assign the first variables ***
        # The frame that contains the ListboxEditable
        self.frameMaster=frameMaster
        # List of the initial items
        self.list=lista
        # Number of initial rows at the moment
        self.numberRows=len(self.list)

        self.placed=False
        self.selected=1
        # *** Create the necessary labels ***
        ind=1
        for row in self.list:
            # Get the name of the label
            labelName='label'+str(ind)
            # Create the variable
            setattr(self, labelName, Label(self.frameMaster, text=self.list[ind-1], bg=colorActiveTab, fg='black', font=(fontLabels, sizeLabels2), pady=2, padx=2, width=20))

            # ** Bind actions
            # 1 left click - Change background
            getattr(self, labelName).bind('<Button-1>',lambda event, a=labelName, ind=ind: self.changeBackground(a, ind))
            # Double click - Convert to entry
            getattr(self, labelName).bind('<Double-1>',lambda event, a=ind: self.changeToEntry(a))
            # Move up and down
            getattr(self, labelName).bind("<Up>",lambda event, a=ind: self.up(a))
            getattr(self, labelName).bind("<Down>",lambda event, a=ind: self.down(a))

            # Increase the iterator
            ind=ind+1

    # Place
    def placeListBoxEditable(self):
        ind=1
        self.placed=True
        # Go row by row placing it
        for row in self.list:
            # Get the name of the label
            labelName='label'+str(ind)
            # Place the variable
            getattr(self, labelName).grid(row=ind-1,column=0)

            # Increase the iterator
            ind=ind+1

    def UpdateListBoxEditable(self,lista):
        ind=1
        self.list=lista
        for row in self.list:
            # Get the name of the label
            labelName='label'+str(ind)
            if self.placed:
                try:
                    getattr(self, labelName).grid_remove()
                except:
                    pass
            # Create the variable
            setattr(self, labelName, Label(self.frameMaster, text=self.list[ind-1], bg=colorActiveTab, fg='black', font=(fontLabels, sizeLabels2), pady=2, padx=2, width=20))

            # ** Bind actions
            # 1 left click - Change background
            getattr(self, labelName).bind('<Button-1>',lambda event, a=labelName: self.changeBackground(a))
            # Double click - Convert to entry
            getattr(self, labelName).bind('<Double-1>',lambda event, a=ind: self.changeToEntry(a))
            # Move up and down
            getattr(self, labelName).bind("<Up>",lambda event, a=ind: self.up(a))
            getattr(self, labelName).bind("<Down>",lambda event, a=ind: self.down(a))

            # Increase the iterator
            ind=ind+1

        self.placeListBoxEditable()
        self.changeBackground('label'+str(self.selected))



    # Action to do when one click
    def changeBackground(self,labelNameSelected):
        # Ensure that all the remaining labels are deselected
        ind=1
        for row in self.list:
            # Get the name of the label
            labelName='label'+str(ind)
            # Place the variable
            try:
                getattr(self, labelName).configure(bg=colorActiveTab)
            except:
                pass
            # Increase the iterator
            ind=ind+1
        # Change the background of the corresponding label
        self.selected=int(labelNameSelected[5:])
        getattr(self, labelNameSelected).configure(bg=colorNoActiveTab)
        # Set the focus for future bindings (moves)
        getattr(self, labelNameSelected).focus_set()


    # Function to do when up button pressed
    def up(self, ind):
        if ind==1: # Go to the last
            # Get the name of the label
            labelName='label'+str(len(self.list))
            self.selected=self.numberRows
        else: # Normal
            # Get the name of the label
            labelName='label'+str(ind-1)
            self.selected=ind-1

        # Call the select
        self.changeBackground(labelName)


    # Function to do when down button pressed
    def down(self, ind):
        if ind==len(self.list): # Go to the last
            # Get the name of the label
            labelName='label1'
            self.selected=1
        else: # Normal
            # Get the name of the label
            labelName='label'+str(ind+1)
            self.selected=ind+1

        # Call the select
        self.changeBackground(labelName)


    # Action to do when double-click
    def changeToEntry(self,ind):
        # Variable of the current entry
        self.entryVar=StringVar()
        # Create the entry
        #entryName='entry'+str(ind) # Name
        self.entryActive=ttk.Entry(self.frameMaster, font=(fontLabels, sizeLabels2), textvariable=self.entryVar, width=200)
        # Place it on the correct grid position
        self.entryActive.grid(sticky="we",row=ind-1,column=0)
        # Focus to the entry
        self.entryActive.focus_set()

        # Bind the action of focusOut
        self.entryActive.bind("<FocusOut>",lambda event, a=ind: self.saveEntryValue(a))

    
    # Action to do when focus out from the entry
    def saveEntryValue(self,ind):
        # Find the label to recover
        labelName='label'+str(ind)
        # Remove the entry from the screen
        self.entryActive.grid_forget()
        # Place it again
        getattr(self, labelName).grid(row=ind-1,column=0)
        # Change the name to the value of the entry
        getattr(self, labelName).configure(text=self.entryVar.get())


    def get_selected(self):
        return self.selected-1


import os, sys, webbrowser, platform
import signal
import threading
import time
import traceback
import re
import pandas as pd
import urllib
from pathlib import Path
from datetime import datetime
import torch
import utm
import matplotlib.pyplot as plt
import numpy as np
from math import radians, cos, sin, sqrt, atan2
import cv2
from sqlalchemy import create_engine, text
from submodules.GPModels import GaussianProcessGPyTorch
from submodules.MQTT import MQTT
import json
import mysql.connector
import pymysql
pymysql.install_as_MySQLdb()
from time import sleep

# import machine this is for hard reset (machine.reset())
version = "1.0.0"
max_tab_idx = -1

class mysql_hdlr():
    def __init__(self,database):
        self.database=database
        self.table = "wqp_gaussian_maps"
        connect_str=f"mysql+mysqlconnector://root:{urllib.parse.quote_plus('azkenmugaÑ.')}@golem.us.es:6006/alec?charset=utf8mb4"
        # print(connect_str)        
        self.motor_mysql = create_engine(connect_str)

    def read_table(self, table):
        self.table=table
        try:
            with self.motor_mysql.connect() as conn:
                self.df = pd.read_sql_table(table, conn)
                self.dft=self.df.T
                print(f"table {self.database}.{self.table} loaded correctly")
        except:
            err=traceback.format_exc()
            print(f"table {self.database}.{self.table} is broken\n {err}")
            return
        
    def increase_max_message_size(self):
        cnx = mysql.connector.connect(user="root", password="azkenmugaÑ.", port="6006",
                                    host="golem.us.es", database="alec")
        # Crear un cursor
        cursor = cnx.cursor()

        # Ejecutar la consulta
        try:
            cursor.execute('SET GLOBAL max_allowed_packet=67108864')
        except:
            err=traceback.format_exc()
            print(f"limit augment failed: \n{err}")
            raise
        
        
    def query_db(self, query=None, database = "wqp"): #Query can also be Positions
        cnx = mysql.connector.connect(user="root", password="azkenmugaÑ.", port="6006",
                                    host="golem.us.es", database=database)
        # Crear un cursor
        cursor = cnx.cursor()

        # Ejecutar la consulta
        try:
            cursor.execute(query)
        except:
            err=traceback.format_exc()
            print(f"query {query} failed: \n{err}")
            raise
        
        # Obtener nombres de columnas
        column_names = cursor.column_names

        # Obtener los resultados
        results = cursor.fetchall()

        # Cerrar el cursor y la conexión a la base de datos
        cursor.close()
        cnx.close()

        aux_df = pd.DataFrame(results, columns=column_names)

        # Crear un DataFrame de pandas con los resultados

        if len(aux_df)==0:
            print(f"empty dataframe at {query}")
        else:
            print(f"queried {len(aux_df)} rows")

        return aux_df

        
    def load_df(self,df):
        self.df=df
        self.dft=df.T
        
    
    def overwrite(self, method = "append"):
        if method == "replace":
            print(f"start overwrite of table {self.table}")
        elif method == "append":
            print(f"adding new data to {self.table}")
        with self.motor_mysql.connect() as conn:
            self.df.to_sql(self.table, conn, if_exists=method, index=False)
        print("process done")

    def cleanup(self):
        print(f"cleaning table {self.table}")
        with self.motor_mysql.connect() as conn:
            with conn.begin():
                #borrar toda la tabla
                conn.execute(text(f"DELETE FROM {self.table}"))



class ACETI_SERVER():
    def __init__(self):
        #init SQL engine
        self.sql = mysql_hdlr("wqp_gaussian_maps")
        #init MQTT engine
        topics=["#"]
        self.mqtt= MQTT(on_message=self.on_message, 
                        topics2suscribe=topics)
        #init
        # Create folder path for data
        Path("data/GP_map/").mkdir(parents=True, exist_ok=True)

        self.sensors=['Conductivity', 'PH', 'Sonar', 'Temperature', 'Turbidity']
        
        mapas=["Alamillo", "Alamillo30x49", "AlamilloAccess11x15"]



        #para cada día y sensor generar un mapa
        days=self.sql.query_db(query= f"SELECT distinct DATE_FORMAT(Date, '%Y-%m-%d') FROM wqp.WQP WHERE Latitude<>0 AND Sensor='Conductivity' AND Data<>0;")
        #for each day
        first=True
        self.sql.cleanup()
        for i in days.values:    
            for mapa in mapas:        
                if mapa=="Alamillo":
                    self.scenario_map = np.genfromtxt(f"assets/Maps/alamillo.csv", delimiter=",")
                    self.map_coords = { 'lat_min': 37.417823087, 'lat_max': 37.421340387, 'lon_min': -6.001553482, 'lon_max': -5.997342323 }
                else:
                    self.latlonmap = np.load(resource_path(f"assets/Maps/{mapa}latlon.npy"), allow_pickle=True)
                    self.scenario_map = np.load(resource_path(f"assets/Maps/{mapa}mask.npy"), allow_pickle=True)
                    self.map_coords = { 'lat_min': self.latlonmap[-1,-1,0], 'lat_max': self.latlonmap[0,0,0], 'lon_min': self.latlonmap[0,0,1], 'lon_max': self.latlonmap[-1,-1,1] }

                self.pos_ini = utm.from_latlon(self.map_coords['lat_min'], self.map_coords['lon_min'])

                #get map shape
                self.rows, self.cols = self.scenario_map.shape
                self.res_lat, self.res_lon = (self.map_coords['lat_max'] - self.map_coords['lat_min']) / self.rows, (self.map_coords['lon_max'] - self.map_coords['lon_min']) / self.cols
                aux_df=self.generate_maps(self.sql.query_db(query=f"SELECT * FROM wqp.WQP where date(Date) = '{i[0]}' AND Latitude<>0 AND Longitude<>0;", database="wqp"))
                aux_df = aux_df.assign(Date=i[0])
                aux_df = aux_df.assign(mapa=mapa)
                if False:
                    #crear un database auxiliar con meanmap y sensor
                    aux_df2=aux_df[aux_df["mean_map"].notnull()][["mean_map", "sensor"]]
                    #guardar aux_df2 en un csv
                    aux_df2.to_csv(f"data/GP_map/{i[0]}.csv", index=False)
                    exit()
                max_size=200 #we will send packet by packet
                for j in range(np.ceil(len(aux_df)/max_size).astype(int)):
                    transmitable_df=aux_df[j*max_size:min((j+1)*max_size, len(aux_df))]
                    self.sql.load_df(transmitable_df)
                    if first:
                        first=False
                        self.sql.overwrite(method="replace")
                    else:
                        self.sql.overwrite(method="append")
                # transmitable_df=aux_df[(len(aux_df)-max_size):]
                # self.sql.load_df(transmitable_df)
                # self.sql.overwrite(method="append")

            


    
    def generate_maps(self, df, rewrite=False):
        # Convertir las coordenadas GPS a metros respecto al inicio de la imagen con la función haversine
        df_=pd.DataFrame()
        for sensor in self.sensors:
            aux_df=pd.DataFrame()
            sensor_df = df[df['Sensor'] == sensor].reset_index(drop=True)
            
            # Quitar las medidas con valor cero
            sensor_df = sensor_df[sensor_df['Data'] != 0].reset_index(drop=True)
            sensor_df=sensor_df.drop(columns=["Sensor"])

            # Limpiar outliers
            Q1 = sensor_df['Data'].quantile(0.25)
            Q3 = sensor_df['Data'].quantile(0.75)
            IQR = Q3 - Q1
            sensor_df = sensor_df[~((sensor_df['Data'] < (Q1 - 3 * IQR)) | (sensor_df['Data'] > (Q3 + 3 * IQR)))]

            sensor_df["Latitude"] = np.ceil((sensor_df["Latitude"].astype(float).values* 10000))/10000 -0.00005
            sensor_df["Longitude"] = np.ceil((sensor_df["Longitude"].astype(float).values* 10000))/10000 -0.00005

            positions = sensor_df.groupby(['Latitude', 'Longitude']).mean()

            y, x = zip(*[self.gps_to_matrix_idx(lat, lon, self.map_coords['lat_max'], self.map_coords['lon_min'], self.res_lat, self.res_lon) for lat, lon in positions.index.values])
            xy = list(zip(y, x)) # Pares de coordenadas GPS
            aux_df["x"]=x
            aux_df["y"]=y
            aux_df["x"]=aux_df["x"].astype(int)
            aux_df["y"]=aux_df["y"].astype(int)


            data = [float(positions["Data"].loc[row]) for row in positions.index.values]
            data=np.array(data)

            if sensor == 'Sonar':# Pasar a metros
                data = data / 1000

            gaussian_process = GaussianProcessGPyTorch(scenario_map = self.scenario_map, initial_lengthscale = 300, kernel_bounds = (200, 400), training_iterations = 50, scale_kernel=True, device = 'cuda' if torch.cuda.is_available() else 'cpu', mean = data.mean())
            gaussian_process.fit_gp(X_new=xy, y_new=data, variances_new=[0.005]*len(data))

            mean_map, uncertainty_map = gaussian_process.predict_gt()
            list_mean_map =  mean_map.tolist()
            str_mean_map = map(str, list_mean_map)
            aux_df=pd.concat([aux_df, pd.DataFrame(data={"mean_map" : str_mean_map})], axis=1)

            # aux_df=pd.concat([aux_df, pd.DataFrame(data={"uncertainty_map" : np.asarray(uncertainty_map).flatten()})], axis=1)
            aux_df=aux_df.assign(sensor= sensor)
            # self.create_map(mean_map, uncertainty_map, x, y, sensor=sensor)

            df_=pd.concat([df_, aux_df], ignore_index = True)

        return df_

    def create_map(self, mean_map, uncertainty_map, x, y, sensor="Conductivity"):

        plt.clf()
        # fig, axis = plt.subplots()
        # Punto de despliegue
        # plt.text(350, 1100, 'Punto de despliegue', fontsize=9, rotation=0, ha='center', va='center', color='w')
        # plt.scatter(175, 1050, c='r', s=50, marker='X', zorder=2)
        plt.xticks([])
        plt.yticks([])

        # Contorno
        cs_internos = plt.contour(mean_map, colors='black', alpha=0.7, linewidths=0.7, zorder=1)
        cs_externo = plt.contour(mean_map, colors='black', alpha=1, linewidths=1.7, zorder=1)

        cs_internos.collections[0].remove()
        for i in range(1, len(cs_externo.collections)):
            cs_externo.collections[i].remove()
        plt.clabel(cs_internos, inline=1, fontsize=3.5)

        # Mapa y puntos de muestreo
        plt.scatter(x, y, c='black', s=1, marker='.', alpha=0.5)
        # vmin_dict = {'Sonar': 2, 'Conductivity': 2.29, 'PH': 7.48, 'Temperature': 17.1, 'Turbidity': 30}
        # vmax_dict = {'Sonar': 0.5, 'Conductivity': 2.14, 'PH': 7.16, 'Temperature': 14.50, 'Turbidity': 15}
        # plt.imshow(mean_map, cmap='viridis', alpha=1, origin='upper', vmin=vmin_dict[sensor], vmax=vmax_dict[sensor])
        vmin = np.min(mean_map[mean_map > 0])
        vmax = np.max(mean_map[mean_map > 0])
        plt.imshow(mean_map, cmap='viridis', alpha=1, origin='upper', vmin=vmin, vmax=vmax)

        # Recortar el mapa
        # plt.ylim(1150, 200)

        # Leyendas
        unidades_dict = {'Sonar': 'Profundidad (m)', 'Conductivity': 'Conductividad (mS/cm)', 'PH': 'pH', 'Temperature': 'Temperatura (ºC)', 'Turbidity': 'Turbidez (NTU)'}
        nombre_dict = {'Sonar': 'Batimetría', 'Conductivity': 'Conductividad', 'PH': 'pH', 'Temperature': 'Temperatura', 'Turbidity': 'Turbidez'}
        plt.colorbar(shrink=0.65).set_label(label=unidades_dict["Conductivity"],size=12)#,weight='bold')
        # plt.text(1950, 650, unidades_dict[sensor], fontsize=12, rotation=90, ha='center', va='center', color='k')
        plt.title(f'{nombre_dict[sensor]} del Lago Mayor (Parque del Alamillo)')
        # if(not os.path.exists("outs")):
        #     os.makedirs("outs")
        # savepath=self.resource_path(f"outs/{nombre_dict[sensor]}_{self.selected_map}.pdf")
        # plt.savefig(savepath, format='pdf')
        plt.show()


    def mainloop(self):
        # query=f"SELECT * FROM wqp.WQP where date(Date) = '{date_}' AND Latitude<>0 AND Sensor='{self.taget_sensor}';"
        sleep(2)



    def dummy(self):
        pass

    def on_message(self, _client, user_data, msg):
        try:
            if "asv_state" in msg.topic:
                self.last_message=json.loads(msg.payload)
            elif "database/wqp" in msg.topic:
                pass
        except:
            error= traceback.format_exc()
            print(f"There was an error parsing mqtt \n{error}")


    def haversine(self, lat1, lon1, lat2, lon2):
        # Radio de la Tierra en kilómetros
        R = 6371.0
        
        # Convertir coordenadas de grados a radianes
        lat1_rad = radians(lat1)
        lon1_rad = radians(lon1)
        lat2_rad = radians(lat2)
        lon2_rad = radians(lon2)
        
        # Diferencias de coordenadas
        dlon = lon2_rad - lon1_rad
        dlat = lat2_rad - lat1_rad
        
        # Fórmula del haversine
        a = sin(dlat / 2)**2 + cos(lat1_rad) * cos(lat2_rad) * sin(dlon / 2)**2
        c = 2 * atan2(sqrt(a), sqrt(1 - a))
        
        # Distancia total en kilómetros
        distance = R * c * 1000
        return distance
    
    def gps_to_matrix_idx(self, lat, lon, lat_max, lon_min, res_lat, res_lon):

        row_idx = int((lat_max - lat) / res_lat)
        col_idx = int((lon - lon_min) / res_lon)

        # Limitar los índices dentro de los rangos válidos de la matriz
        # row_idx = max(0, min(row_idx, rows - 1))
        # col_idx = max(0, min(col_idx, cols - 1))

        return row_idx, col_idx


def resource_path(relative_path):
        """ Get absolute path to resource, works for dev and for PyInstaller """
        try:
            # PyInstaller creates a temp folder and stores path in _MEIPASS
            base_path = sys._MEIPASS
        except Exception:
            base_path = os.path.abspath(".")

        return os.path.join(base_path, relative_path)


def main(args=None):
    server=ACETI_SERVER()
    print(f"Init Version {version}")
    server.mainloop() 
    print("Finishing Application")
    server.mqtt.close()
    print("Application Finished")
    
if __name__ == '__main__':
    main()



import tkinter
import tkinter.ttk
import tkinter.font
import threading
import math
import numpy as np
from PIL import Image, ImageTk
from ...common_maps.map_creator import MapDownloader
class MapCreationWindow(tkinter.Toplevel):
    """Minimal Toplevel window to be filled by the user later.

    Usage:
        win = MapCreationWindow(parent)
        win.show()
    """
    def __init__(self, parent, title="Map Creator", width=800, height=600, northwest=[0,0], southeast=[0,0], gridsize=1):
        super().__init__(parent)
        self.parent = parent
        self.title(title)
        self.geometry(f"{width}x{height}")
        self.transient(parent)
        self.grab_set()

        # basic layout
        self.main_frame = tkinter.Frame(self)
        self.main_frame.pack(fill="both", expand=True, padx=8, pady=8)

        self.header = tkinter.Label(self.main_frame, text=title, font=tkinter.font.Font(weight='bold', size=14))
        self.header.pack(side="top", pady=(0,8))

        # placeholder area for future widgets
        self.placeholder = tkinter.Label(self.header, text="Downloading map...")
        self.placeholder.pack(side="top",fill="both", expand=True)
        

        self.gridsize = gridsize

        # initialize cell/grid related state to avoid callback AttributeError
        self._cell_items = {}      # map (row,col) -> canvas item id for overlays
        self._cell_active = {}     # map (row,col) -> bool whether dark overlay is active
        self._cell_draw_params = None
        # selection holders used by click handlers
        self._selected_cell = None
        self._selected_cell2 = None
        # canvas used for cell drawing; some helper methods expect self.cell_canvas
        self.canvas = None
        self.cell_canvas = None

        desired_zoom = math.log2((156543.03392 * math.cos(math.radians((northwest[0] + southeast[0]) / 2))) / (self.gridsize)) + 2
        if desired_zoom > 22:
            desired_zoom = 21
        elif desired_zoom < 0:
            desired_zoom = 0
        print(f"to have {self.gridsize} pixels per grid, you need to set zoom to {int(desired_zoom)}")

        if northwest[0]<southeast[0]:
            aux=northwest[0]
            northwest[0]=southeast[0]
            southeast[0]=aux
        if northwest[1]>southeast[1]:
            aux=northwest[1]
            northwest[1]=southeast[1]
            southeast[1]=aux


        self.mapa = MapDownloader(northwest, southeast, zoom=int(desired_zoom), layer="s", server ="arcgis")

        self.tilenumber = abs(self.mapa.nw_tile[0]-self.mapa.se_tile[0]) * abs(self.mapa.nw_tile[1]-self.mapa.se_tile[1])

        #crea una barra de carga de elements divisiones
        self.progress_bar = tkinter.ttk.Progressbar(self.main_frame, mode="determinate", length= 400, maximum=self.tilenumber)
        self.progress_bar.pack(side="top", fill="x", pady=(8,0))
        self.download_thread = threading.Thread(target=self.start_download)
        self.download_thread.start()
        # action buttons
        buttons = tkinter.Frame(self.main_frame)
        buttons.pack(side="bottom", fill="x", pady=(8,0))

        self.cancel_button = tkinter.Button(buttons, text="Cancel", command=self.on_cancel)
        self.cancel_button.pack(side="right")
        self.bind("<Escape>", self.end_program)



    def start_download(self):
        downloader = self.mapa.image_generator()
        # update progress in the Tk thread via after to avoid thread-safety issues
        for _ in downloader:
            try:
                self.after(0, self.progress_bar.step, 1)
            except Exception:
                # fallback to direct call if after fails for any reason
                try:
                    self.progress_bar.step(1)
                except Exception:
                    pass
        self.progress_bar.destroy()
        self.cancel_button.destroy()
        self.placeholder.config(text="Map download complete.")
        self.after(10, self.create_navigation_map)

    def create_navigation_map(self):
        self.latlonmap = self.mapa.create_grid_latlon(self.gridsize)
        self.navigation_map = np.ones_like(self.latlonmap, dtype=np.uint8)
        self.placeholder.config(text="Map and grid created.")

        self.tools_frame = tkinter.Frame(self.main_frame)
        self.tools_frame.pack(side="right", fill="y", padx=(8,0))

        def _validate_int(self, proposed=0):
            """Return True if proposed string is empty or a float with up to 2 decimals."""
            if proposed == "":
                return True
            try:
                m = int(proposed)
                return m is not None
            except Exception:
                return False

        tkinter.Label(self.tools_frame, text="Pincel Size (cells):").pack(side="left", padx=(0, 10))
        self.pinzel_size = tkinter.StringVar()
        self.pinzel_size.set("1")
        vcmd = (self.register(_validate_int), '%P')
        self.pincel_check = tkinter.Entry(self.tools_frame, textvariable=self.pinzel_size)
        self.pincel_check.pack(side="left", padx=(0, 10))
        self.pincel_check.config(validate='key', validatecommand=vcmd)
        self._cell_active = getattr(self, '_cell_active', {})

        #boton de guardar
        self.save_button = tkinter.Button(self.tools_frame, text="Save Navigation Map", command=self.save_navigation_map)
        self.save_button.pack(side="top", pady=(8,0))

        # create a canvas that expands with the window and will display a rescaled map
        if hasattr(self, 'canvas') and self.canvas:
            try:
                # remove previous image item if any
                if hasattr(self, '_image_id'):
                    try:
                        self.canvas.delete(self._image_id)
                    except Exception:
                        pass
            except Exception:
                pass

        self.canvas = tkinter.Canvas(self.main_frame, bg='white')
        self.canvas.pack(side="top", fill="both", expand=True)

        # keep a copy of the original PIL image (we will resize from this on demand)
        try:
            self._orig_map_img = self.mapa.map_img.copy()
        except Exception:
            self._orig_map_img = self.mapa.map_img

        # create initial PhotoImage (will be updated on first configure)
        try:
            self.tk_map_img = ImageTk.PhotoImage(self._orig_map_img)
        except Exception:
            try:
                img = self._orig_map_img.convert('RGBA')
                self.tk_map_img = ImageTk.PhotoImage(img)
            except Exception as e:
                print(f"Failed to create PhotoImage from map image: {e}")
                return

        # create image on canvas and store its id (will be repositioned on resize)
        self.canvas.create_image(0, 0, anchor="nw", image=self.tk_map_img)


        # store handler so it can be unbound if needed
        self.canvas.bind('<Configure>', self._on_canvas_config)

        self.canvas.bind('<B1-Motion>', self._on_cell_click)
        self.canvas.bind('<B3-Motion>', self._on_cell_right_click)

        # bind resize handler to rescale the image while keeping aspect ratio
    def _on_canvas_config(self, event=None):
        canvas = self.canvas
        # nothing to do if image/grid not ready
        if not hasattr(self, '_orig_map_img') or not hasattr(self, 'latlonmap'):
            return

        # get canvas size
        c_w = max(canvas.winfo_width(), 2)
        c_h = max(canvas.winfo_height(), 2)

        img_w, img_h = self._orig_map_img.size
        # compute scale to fit image inside canvas while keeping aspect ratio
        scale = min(c_w / img_w, c_h / img_h)
        new_w = max(1, int(img_w * scale))
        new_h = max(1, int(img_h * scale))

        # resize original image to new size
        try:
            resized = self._orig_map_img.resize((new_w, new_h), resample=Image.LANCZOS)
        except Exception:
            resized = self._orig_map_img.copy()

        # create a new PhotoImage and keep reference to avoid GC
        try:
            self.tk_map_img = ImageTk.PhotoImage(resized)
        except Exception:
            try:
                self.tk_map_img = ImageTk.PhotoImage(resized.convert('RGBA'))
            except Exception as e:
                print(f"Failed to create resized PhotoImage: {e}")
                return

        # clear canvas and draw the resized image at top-left (0,0)
        canvas.delete('all')
        self._image_id = canvas.create_image(0, 0, anchor='nw', image=self.tk_map_img)

        # compute grid cell size from the resized image and the latlonmap shape
        w, h = self.latlonmap.shape
        print(f"latlonmap shape: {h} rows x {w} cols, scale: {scale:.5f} image size: {new_w}x{new_h} old {img_w}x{img_h} active cells: {sum(self._cell_active.values())}")
        cell_w = max(1, new_w / w)
        cell_h = max(1, new_h / h)

        # draw rectangles and remember their ids
        self._cell_items = getattr(self, '_cell_items', {})
        self._cell_items.clear()
        for r in range(h):
            y1 = r * cell_h
            y2 = y1 + cell_h
            for c in range(w):
                x1 = c * cell_w
                x2 = x1 + cell_w
                tag = f"cell_{r}_{c}"
                rid = canvas.create_rectangle(x1, y1, x2, y2, outline='lightgrey', width=1, tags=(tag, 'cell'))
                self._cell_items[(r, c)] = rid
                try:
                    if self._cell_active[(r, c)]:
                        self._highlight_cell(r, c, color='gray')
                        print(f"cell {(r,c)} is active", end='')
                except:
                    self._cell_active[(r, c)] = False

        # resize canvas scrollregion
        canvas.config(scrollregion=(0, 0, w * cell_w, h * cell_h))
        # ensure references for other methods
        self.cell_canvas = canvas
        # save draw params
        self._cell_draw_params = dict(cell_w=cell_w, cell_h=cell_h, rows=h, cols=w, img_w=new_w, img_h=new_h, scale=scale)

    def _on_cell_click(self, event):
        """Handle click on canvas: compute row/col, highlight and process."""
        if not hasattr(self, '_cell_draw_params'):
            return
        params = self._cell_draw_params
        _row = int(event.y / params['cell_h'])
        _col = int(event.x / params['cell_w'])
        for row in range(_row - int(self.pinzel_size.get()), _row + int(self.pinzel_size.get())+1):
            for col in range(_col - int(self.pinzel_size.get()), _col + int(self.pinzel_size.get())+1):
                # bounds check
                if row < 0 or row >= params['rows'] or col < 0 or col >= params['cols']:
                    continue
                # highlight new
                self._highlight_cell(row, col, color='gray')
                # process selection: update UI labels
                try:
                    self.navigation_map[row, col] = 0 # mark as non navigable
                except Exception:
                    pass
                self._cell_active[(row, col)] = True

    def _on_cell_right_click(self, event):
        """Handle right click on canvas: compute row/col, highlight and process."""
        if not hasattr(self, '_cell_draw_params'):
            return
        params = self._cell_draw_params

        _row = int(event.y / params['cell_h'])
        _col = int(event.x / params['cell_w'])
        for row in range(_row - int(self.pinzel_size.get()), _row + int(self.pinzel_size.get())+1):
            for col in range(_col - int(self.pinzel_size.get()), _col + int(self.pinzel_size.get())+1):
                # bounds check
                if row < 0 or row >= params['rows'] or col < 0 or col >= params['cols']:
                    continue
                # highlight new
                self._remove_highlight(row, col)
                # process selection: update UI labels
                try:
                    self.navigation_map[row, col] = 1 # mark as non navigable
                except Exception:
                    pass
                self._cell_active[(row, col)] = False


    def _highlight_cell(self, row, col, color='red'):
        if self._cell_active[(row, col)]:
            return  # already active
        canvas = self.canvas
        params = self._cell_draw_params
        sw = params['cell_w']
        sh = params['cell_h']

        x1 = col * sw
        y1 = row * sh
        x2 = x1 + sw
        y2 = y1 + sh
        # draw an outline rectangle on top with tag 'sel'
        canvas.create_rectangle(x1, y1, x2, y2, fill=color, width=2, tags=('sel', f'sel_{row}_{col}'))

    def _remove_highlight(self, row, col):
        if not self._cell_active[(row, col)]:
            return  # already inactive
        canvas = self.canvas
        tag = f'sel_{row}_{col}'
        canvas.delete(tag)

    def on_cancel(self):
        self.destroy()

    def show(self):
        self.wait_window(self)

    def end_program(self, event=None):
        if self.download_thread.is_alive():
            self.download_thread.join(timeout=10)
        self.on_cancel()

    def save_navigation_map(self):
        import tkinter.filedialog

        file_path = tkinter.filedialog.asksaveasfilename(
            parent=self,
            title="Save Navigation Map",
            filetypes=[("Navigation Map Files", "*.npy"), ("All Files", "*.*")],
        )
        if not file_path:
            return  # user cancelled
        
        if ".npy" not in file_path:
            file_path += ".npy"
        np.save(file_path.replace('.npy', 'mask.npy'), self.navigation_map.T)
        np.save(file_path.replace('.npy', 'latlon.npy'), self.latlonmap)



from aceti.common_maps import import_map, MapDownloader
import numpy as np
import matplotlib.pyplot as plt
import PIL

def plot_grid(map):
    ''' Plot a grid on top of a map. '''

    #check map properties
    if isinstance(map, str):
        base_matrix, lat_lon_map  = import_map(map)
    elif isinstance(map, tuple):
        base_matrix, lat_lon_map  = map
        if isinstance(base_matrix, np.ndarray) and isinstance(lat_lon_map, np.ndarray):
            if not np.all(np.isin(base_matrix, [0, 1])):
                raise ValueError("base_matrix must be a numpy array of 0s and 1s")
            if base_matrix.shape != lat_lon_map.shape:
                raise ValueError("base_matrix and lat_lon_map must have the same shape")
            if not np.all(np.apply_along_axis(lambda x: isinstance(x, tuple) and len(x) == 2 and all(isinstance(i, float) for i in x), 1, lat_lon_map)):
                raise ValueError("lat_lon_map must be a numpy array of tuples of 2 coordinates, [lat, lon]")
        else:
            raise ValueError("map must be a string or a tuple of [base_matrix, lat_lon_map]")
    else:
        raise ValueError("map must be a string or a tuple of [base_matrix, lat_lon_map]")
    # Plot the map
    print("lat_lon_map shape: ", lat_lon_map.shape)
    print("base_matrix shape: ", base_matrix.shape)
    print("lat_lon_map: ", lat_lon_map[0, 0])
    print("lat_lon_map: ", lat_lon_map[-1, -1])
    #to tuple
    northwest = [float(lat_lon_map[0, 0][0]), float(lat_lon_map[0, 0][1])]
    southeast = [float(lat_lon_map[-1, -1][0]), float(lat_lon_map[-1, -1][1])]
    map=MapDownloader(northwest, southeast, zoom=19, layer="s", server ="arcgis") 
    map.generateImage()
    # map.raw_map_img = PIL.Image.open("AlamilloPlantilla.png")
    # plt.imshow(map.map_img)
    # plt.savefig("alamillo95x216.png", dpi=600)
    map.show_grid(lat_lon_map, base_matrix)

    # plt.figure()
    # plt.imshow(base_matrix, cmap='gray', interpolation='nearest')
    # plt.show()

    


# csv_to_npy("alamillo95x216")
# csv_to_png("alamillo95x216plantilla")

# plot_grid("alamilloaccess11x15")
# plot_grid("alamillo95x216")
plot_grid("/home/aloepacci/guadaira")
# plot_grid("alamillo30x49")
# map=MapDownloader([37.191611287951396, -4.399097018031518], [37.188098777474906, -4.394347328263865], zoom=19, layer="s", server ="arcgis")
# map=MapDownloader([37.341664356846564, -5.858012026866486], [37.33741741302848, -5.855613536856963], zoom=19, layer="s", server ="arcgis")
# map.save_navigation_waters(name="guadaira")

# plt.imshow(map.map_img)
# plt.axis('off')
# plt.savefig("guadaira.png", dpi=600)





import numpy as np
import os
import sys
from matplotlib import pyplot as plt


if sys.version_info < (3, 9):
    # importlib.resources either doesn't exist or lacks the files() function, so use the PyPI version:
    import importlib_resources
else:
    # importlib.resources has files(), so use that:
    import importlib.resources as importlib_resources
pkg = importlib_resources.files("aceti")

def import_map(map_name: str):
    """
    Import a map from a file and return the base matrix and lat/lon map.
    Args:
        map_name (str): The name of the map to import.
    Returns:
        tuple: A tuple containing the base matrix and lat/lon map.
    """
    map_name = find_map(map_name)

    try:
        base_matrix =  np.load(f"{map_name}mask.npy")
    except FileNotFoundError:
        raise ValueError(f"No binnary mask found in package directory {map_name}mask.npy")
    # Load the lat/lon map
    try:
        lat_lon_map = np.load(f"{map_name}latlon.npy", allow_pickle=True)
    except FileNotFoundError:
        raise ValueError(f"No latlon mask found in package directory {map_name}latlon.npy")
    
    base_matrix = base_matrix.astype(int)
    return base_matrix, lat_lon_map


def plot_map(map_name):
    ''' Convert a CSV file to a binary image. '''
    map_name = find_map(map_name)
    binary_image = np.load(pkg.joinpath("maps", f"{map_name}mask.npy"))

    # Convert the binary image to 0s and 255s
    binary_image = np.where(binary_image > 0, 255, 0)

    plt.imshow(binary_image, cmap='gray')
    plt.axis('off')
    plt.show()

def map_downsize(map_name, factor=2):
    ''' Downsize a map. '''
    map_name = find_map(map_name)
    binary_image = np.load(pkg.joinpath("maps", f"{map_name}mask.npy"))
    return binary_image[::factor, ::factor]

def map_shrink(map_name, init_column=0, final_column=None, init_row=0, final_row=None):
    ''' Shrink a map. '''
    map_name = find_map(map_name)
    binary_image = np.load(pkg.joinpath("maps", f"{map_name}mask.npy"))

    # Set the final column and row if they are None
    if final_column is None:
        final_column = binary_image.shape[1]
    if final_row is None:
        final_row = binary_image.shape[0]

    # Shrink the binary image
    return binary_image[init_row:final_row, init_column:final_column]

def plot_grid(map):
    ''' Plot a grid on top of a map. '''
    base_matrix, lat_lon_map  = import_map(map)

    # if isinstance(base_matrix, np.ndarray) and isinstance(lat_lon_map, np.ndarray):
    #     if not np.all(np.isin(base_matrix, [0, 1])):
    #         raise ValueError("base_matrix must be a numpy array of 0s and 1s")
    #     if base_matrix.shape != lat_lon_map.shape:
    #         raise ValueError("base_matrix and lat_lon_map must have the same shape")
    #     if not np.all(np.apply_along_axis(lambda x: isinstance(x, tuple) and len(x) == 2 and all(isinstance(i, float) for i in x), 1, lat_lon_map)):
    #         raise ValueError("lat_lon_map must be a numpy array of tuples of 2 coordinates, [lat, lon]")
    # else:
    #     raise ValueError("map must be a string or a tuple of [base_matrix, lat_lon_map]")
    # Plot the map
    from aceti import MapDownloader
    map=MapDownloader(lat_lon_map[0, 0], lat_lon_map[-1, -1], zoom=19, layer="s", server ="arcgis")  #TODO: assign zoom depending on size
    map.show_grid(lat_lon_map, base_matrix)


def find_map(map_name):
    ''' Find a map in the package directory. '''
    map_name = map_name.lower()
    if os.path.exists(f'{map_name}mask.npy') and os.path.exists(f'{map_name}latlon.npy'):
        print(f"Map {map_name} found ")
        return map_name
    elif os.path.exists(pkg.joinpath("maps", f"{map_name}mask.npy")) and os.path.exists(pkg.joinpath("maps", f"{map_name}latlon.npy")):
        print(f"Map {map_name} found in package directory")
        return pkg.joinpath(f"maps/{map_name}")
    else:
        #check for path
        if os.path.exists(f'{map_name}mask.csv'):
            binary_image = np.loadtxt(f'{map_name}mask.csv', delimiter=" ")
        elif os.path.exists(f'{map_name}plantilla.csv'):
            binary_image = np.loadtxt(f'{map_name}plantilla.csv', delimiter=" ")
        else:
            raise ValueError('map binnary not found')

        if os.path.exists(f'{map_name}latlon.csv'):
            latlon_image = np.loadtxt(f'{map_name}latlon.csv', delimiter=";", dtype=str)
        elif os.path.exists(f'{map_name}grid.csv'):
            latlon_image = np.loadtxt(f'{map_name}grid.csv', delimiter=";", dtype=str)
        else:
            raise ValueError('map latlon not found')
        # Convert to npy
        print("maps found, converting to npy")
        np.save(f'{map_name}mask.npy', binary_image)


        lat_lon_map = np.zeros((binary_image.shape[0], binary_image.shape[1], 2), dtype=np.float64)
        # Convert the strings to tuples
        for i in range(binary_image.shape[0]):
            for j in range(binary_image.shape[1]):
                lat = latlon_image[i, j].split(',')[0]
                lon = latlon_image[i, j].split(',')[1]
                lat_lon_map[i, j] = [float(lat), float(lon)]
        # Save the binary image as a NumPy file
        np.save(f'{map_name}latlon.npy', lat_lon_map)
        return map_name




import tkinter
import tkinter.font
import tkinter.ttk
import os, sys, webbrowser, platform
import signal
import random
from tkintermapview import TkinterMapView
from tkintermapview.canvas_position_marker import CanvasPositionMarker
import threading
import time
import screeninfo
import traceback
import re
import requests
import math
import pandas as pd
from PIL import Image, ImageTk
sys.path.append('../')
from ..shared import *
from ..assets import *
from ..widgets.ListboxEditable import ListboxEditable
from ..widgets.labels import *
from .map_creation_window import MapCreationWindow
max_retain_data=40


class MAP_GENERATOR_WIDGET(TkinterMapView):
    def __init__(self, *args, parent=None, **kwargs):
        self.assets=parent.assets
        self.parent=parent
        super().__init__(*args,**kwargs)


class MAP_GENERATORTAB(tkinter.ttk.PanedWindow):
    def __init__(self, parent=None):
        if parent is None:
            parent=tkinter.Tk()
        self.parent=parent
        self.polygon=None
        #inherit
        try:
            self.assets=self.parent.assets
            self.shared=self.parent.shared
        except:
            self.assets=Assets()
            self.shared=SHARED()

        super().__init__(orient="horizontal")
        self.gs_tab()

    def gs_tab(self):
        self.GPS_panel = tkinter.ttk.PanedWindow(orient="vertical")
        self.add(self.GPS_panel)

        #create map
        self.map_widget = MAP_GENERATOR_WIDGET(parent= self,corner_radius=0,width=int(self.parent.screenwidth*0.75)+1, height=int(self.parent.screenheight*0.75)+1) #this widget has no automatic size
        self.map_widget.set_tile_server("https://mt0.google.com/vt/lyrs=s&hl=en&x={x}&y={y}&z={z}&s=Ga", max_zoom=22)
        self.GPS_panel.add(self.map_widget)
        # self.map_widget.pack_propagate(False)
        self.topleft=[37.420088,-6.001346]
        self.bottomright=[37.418399,-5.997504]
        self.map_widget.fit_bounding_box(self.topleft, self.bottomright)

        #map selection
        self.map_selection=tkinter.Frame(self, width=20)
        self.map_selection.pack_propagate(False)
        self.add(self.map_selection)

        self.map_selection_frame=tkinter.Frame(self.map_selection)
        self.map_selection_frame.pack(side="top", fill="x")
        #Label barra de búsqueda
        self.map_search_label=tkinter.Label(self.map_selection_frame, text="1 Barra de Búsqueda",font=tkinter.font.Font(weight='bold', size=18))
        self.map_search_label.pack(side="top",pady=15)
        # search bar for tkintermapview
        self.search_frame = tkinter.Frame(self.map_selection_frame)
        self.search_entry = tkinter.Entry(self.search_frame)
        self.search_entry.pack(side="left", fill="x", expand=True, padx=(0,4))
        self.search_button = tkinter.Button(self.search_frame, image=self.assets.icon_lupa, command=self.search_location, width=36, height=36)
        self.search_button.pack(side="right")
        self.search_entry.bind("<Return>", self.search_location)
        self.search_frame.pack(side="top", fill="x", pady=(10,5))

        #L selecciona las esquinas del mapa
        self.map_search_label=tkinter.Label(self.map_selection_frame, text="2 Corner selector",font=tkinter.font.Font(weight='bold', size=18))
        self.map_search_label.pack(side="top",pady=15)


        # 2 botones para elegir puntos de coordenadas de dos esquinas cuando pulse en el mapa despues de estos
        self.northwest_selection_frame=tkinter.Frame(self.map_selection)
        self.northwest_selection_frame.pack(side="top", fill="x", pady=(10,5))
        self.map_selection_button1 = tkinter.Button(self.northwest_selection_frame, text="Esquina 1", command=self.select_top_left)
        self.map_selection_button1.pack(side="left", pady=5)
        # use StringVar so we can trace changes
        self.northwest_var_lat = tkinter.StringVar()
        self.northwest_var_lon = tkinter.StringVar()
        self.northwest_entry_lat = tkinter.Entry(self.northwest_selection_frame, textvariable=self.northwest_var_lat)
        self.northwest_entry_lat.pack(side="left", fill="x", expand=True, padx=(0,4))
        self.northwest_entry_lon = tkinter.Entry(self.northwest_selection_frame, textvariable=self.northwest_var_lon)
        self.northwest_entry_lon.pack(side="left", fill="x", expand=True, padx=(0,4))

        self.southeast_selection_frame=tkinter.Frame(self.map_selection)
        self.southeast_selection_frame.pack(side="top", fill="x", pady=(10,5))
        self.map_selection_button2 = tkinter.Button(self.southeast_selection_frame, text="Esquina 2", command=self.select_bottom_right)
        self.map_selection_button2.pack(side="left", pady=5)
        # use StringVar so we can trace changes
        self.southeast_var_lat = tkinter.StringVar()
        self.southeast_var_lon = tkinter.StringVar()
        self.southeast_entry_lat = tkinter.Entry(self.southeast_selection_frame, textvariable=self.southeast_var_lat)
        self.southeast_entry_lat.pack(side="left", fill="x", expand=True, padx=(0,4))
        self.southeast_entry_lon = tkinter.Entry(self.southeast_selection_frame, textvariable=self.southeast_var_lon)
        self.southeast_entry_lon.pack(side="left", fill="x", expand=True, padx=(0,4))

        # add traces so any change to these variables updates the polygon
        try:
            for var in (self.northwest_var_lat, self.northwest_var_lon, self.southeast_var_lat, self.southeast_var_lon):
                # trace_add is preferred in modern tkinter
                try:
                    var.trace_add('write', self.writepolygon)
                except Exception:
                    # fallback to older trace
                    var.trace('w', lambda *args, v=var: self.writepolygon())
        except Exception:
            pass


        #L selecciona las esquinas del mapa
        self.map_config_label=tkinter.Label(self.map_selection_frame, text="3 Configuracion",font=tkinter.font.Font(weight='bold', size=18))
        self.map_config_label.pack(side="top",pady=15)

        # Configuración de la selección del mapa
        self.map_config_frame = tkinter.Frame(self.map_selection_frame)
        self.map_config_frame.pack(side="top", fill="x", pady=(10, 5))

        tkinter.Label(self.map_config_frame, text="Grid Size (meters):").pack(side="left", padx=(0, 10))

        self.gridsize_var = tkinter.StringVar()
        self.gridsize_var.set("1.0")
        vcmd = (self.register(self._validate_float), '%P')
        self.grid_check = tkinter.Entry(self.map_config_frame, textvariable=self.gridsize_var)
        self.grid_check.pack(side="left", padx=(0, 10))
        self.grid_check.config(validate='key', validatecommand=vcmd)
        self.gridsize_var.trace_add('write', lambda *args: print(f"Grid size set to: {self.gridsize_var.get()}"))


        #botón de continuar
        self.map_selection_button = tkinter.Button(self.map_selection_frame, text="Continue", command = self.to_map_creator)
        self.map_selection_button.pack(side="top", pady = 50)

    def select_top_left(self):
        def callback(event):
            try:
                lat, lon = self.map_widget.convert_canvas_coords_to_decimal_coords(event.x, event.y)
            except Exception:
                # fallback: use event coordinates if conversion not available
                try:
                    lat, lon = event.lat, event.lon
                except Exception:
                    print("Could not get coordinates from click event")
                    return
            # write coordinates to entries
            try:
                # update traced variables (these will trigger writepolygon)
                try:
                    self.northwest_var_lat.set(f"{lat:.6f}")
                    self.northwest_var_lon.set(f"{lon:.6f}")
                except Exception:
                    # fallback: write directly to entries
                    self.northwest_entry_lat.delete(0, 'end')
                    self.northwest_entry_lat.insert(0, f"{lat:.6f}")
                    self.northwest_entry_lon.delete(0, 'end')
                    self.northwest_entry_lon.insert(0, f"{lon:.6f}")
            except Exception:
                pass
            # unbind the click handler
            try:
                self.map_widget.canvas.unbind("<Button-1>")
                self.map_widget.canvas.bind("<Button-1>", self.map_widget.mouse_click)

            except Exception:
                pass
        self.map_widget.canvas.unbind("<Button-1>")
        self.map_widget.canvas.bind("<Button-1>", callback)

    def _validate_float(self, proposed):
        """Return True if proposed string is empty or a float with up to 2 decimals."""
        if proposed == "":
            return True
        try:
            # allow a leading minus and decimals
            # but enforce at most 2 decimals while typing
            m = re.match(r'^-?\d*(?:\.(\d{0,2})?)?$', proposed)
            return m is not None
        except Exception:
            return False

    def select_bottom_right(self):
        def callback(event):
            try:
                lat, lon = self.map_widget.convert_canvas_coords_to_decimal_coords(event.x, event.y)
            except Exception:
                try:
                    lat, lon = event.lat, event.lon
                except Exception:
                    print("Could not get coordinates from click event")
                    return
            try:
                # update traced variables (these will trigger writepolygon)
                try:
                    self.southeast_var_lat.set(f"{lat:.6f}")
                    self.southeast_var_lon.set(f"{lon:.6f}")
                except Exception:
                    self.southeast_entry_lat.delete(0, 'end')
                    self.southeast_entry_lat.insert(0, f"{lat:.6f}")
                    self.southeast_entry_lon.delete(0, 'end')
                    self.southeast_entry_lon.insert(0, f"{lon:.6f}")
            except Exception:
                pass
            try:
                self.map_widget.canvas.unbind("<Button-1>")
                self.map_widget.canvas.bind("<Button-1>", self.map_widget.mouse_click)
            except Exception:
                pass
            print(f"Bottom-right corner set to: {self.southeast_entry_lat.get()}, {self.southeast_entry_lon.get()}")

        self.map_widget.canvas.unbind("<Button-1>")
        self.map_widget.canvas.bind("<Button-1>", callback)

    def search_location(self, event=None):
        try:
            query = self.search_entry.get().strip()
            if not query:
                return
            # simple rate-limit to avoid rapid repeated queries
            try:
                import time
                now = time.time()
                if getattr(self, '_last_search', 0) and now - self._last_search < 1.0:
                    # ignore searches less than 1s apart
                    return
                self._last_search = now
            except Exception:
                pass

            # First, try a direct Nominatim HTTP request with proper identification
            try:
                nominatim_url = "https://nominatim.openstreetmap.org/search"
                headers = {"User-Agent": "aceti_maps/0.0.2 (acasado4@us.es)"}
                params = {"q": query, "format": "jsonv2", "addressdetails": 1, "limit": 1, "email": "acasado4@us.es"}
                resp = requests.get(nominatim_url, params=params, headers=headers, timeout=10)
                if resp.status_code == 200:
                    data = resp.json()
                    if data:
                        lat = float(data[0]["lat"])
                        lon = float(data[0]["lon"])
                        try:
                            self.map_widget.set_position(lat, lon)
                            self.map_widget.set_zoom(19)
                            return
                        except Exception:
                            pass
                    else:
                        print(f"No geocoding results for '{query}'")
                else:
                    # Nominatim may return 403 if our requests are blocked/rate-limited
                    print(f"Nominatim HTTP error {resp.status_code} for query '{query}'")
                    if resp.status_code == 403:
                        print("Nominatim returned 403 — you may be rate-limited or blocked. Consider using an API key provider or running a local Nominatim instance.")
            except Exception as e:
                print(f"Direct Nominatim request failed: {e}")

            # Fallback to internal geocoder provided by tkintermapview (may hit same service)
            try:
                self.map_widget.set_address(query)
                return
            except Exception:
                # if that fails, try interpreting the query as 'lat, lon'
                m = re.match(r"\s*([+-]?\d+\.?\d*)[,\s]+([+-]?\d+\.?\d*)\s*$", query)
                if m:
                    lat = float(m.group(1))
                    lon = float(m.group(2))
                    try:
                        self.map_widget.set_position(lat, lon)
                        self.map_widget.set_zoom(15)
                        return
                    except Exception:
                        pass
                print(f"Search failed for '{query}' (no geocoding result)")
        except Exception as e:
            print(f"Error in search_location: {e}")


    def writepolygon(self, *args):
        """Draw or update a rectangular polygon from northwest and southeast variables."""
        try:
            lat1 = float(self.northwest_var_lat.get())
            lon1 = float(self.northwest_var_lon.get())
            lat2 = float(self.southeast_var_lat.get())
            lon2 = float(self.southeast_var_lon.get())
        except Exception:
            return

        # ensure lat1 is the northern (greater) and lat2 is southern (smaller) if user inverted
        try:
            if lat2 > lat1:
                lat1, lat2 = lat2, lat1
            if lon1 > lon2:
                lon1, lon2 = lon2, lon1
        except Exception:
            pass

        coords = [(lat1, lon1), (lat1, lon2), (lat2, lon2), (lat2, lon1), (lat1, lon1)]

        try:
            # remove previous polygon/path if exists
            try:
                if getattr(self, 'polygon', None):
                    try:
                        # some objects may provide a delete method
                        if hasattr(self.polygon, 'delete'):
                            self.polygon.delete()
                        elif hasattr(self.polygon, 'remove'):
                            self.polygon.remove()
                    except Exception:
                        pass
                    self.polygon = None
            except Exception:
                pass

            # try set_polygon (if available), otherwise use set_path
            try:
                self.polygon = self.map_widget.set_polygon(coords)
            except Exception:
                try:
                    # delete all paths to avoid duplicates
                    try:
                        self.map_widget.delete_all_path()
                    except Exception:
                        pass
                    self.polygon = self.map_widget.set_path(coords)
                except Exception as e:
                    print(f"Could not draw polygon: {e}")
        except Exception as e:
            print(f"writepolygon error: {e}")

    def center_map(self,event=None):
        self.map_widget.fit_bounding_box(self.topleft, self.bottomright)

    def to_map_creator(self, event=None):

        try:
            lat1 = float(self.northwest_var_lat.get())
            lon1 = float(self.northwest_var_lon.get())
            lat2 = float(self.southeast_var_lat.get())
            lon2 = float(self.southeast_var_lon.get())
            if lat2 > lat1:
                lat1, lat2 = lat2, lat1
            if lon1 > lon2:
                lon1, lon2 = lon2, lon1
            #if they look like numbers
            if not all(isinstance(coord, float) for coord in (lat1, lon1, lat2, lon2)):
                raise ValueError("Coordinates must be numeric.")
        except Exception as e:
            tkinter.messagebox.showerror("Invalid Coordinates", "Please enter valid numeric coordinates for both corners.")
            return
        try:
            MapCreationWindow(self, northwest=[lat1, lon1], southeast=[lat2, lon2], gridsize=float(self.gridsize_var.get()))
        except Exception as e:
            print(f"Error in to_map_creator: {e}")

    

    def close(self):
        pass

    def open(self):
        pass



# from .maps_utils import png_to_csv, csv_to_png, csv_to_npy, invert_map, get_output_name, csv_shrink, csv_downsize
from .map_importer import import_map, plot_map, map_downsize, map_shrink
from .map_creator import MapDownloader    


from aceti.common_maps import import_map
import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
import math
from tqdm.auto import tqdm
import os
import urllib.request
import pyproj

class MapDownloader:
    """
        A class which generates high resolution google maps image given
        Northwest and Southeast points in [Lat Lon] and zoom level
        You can change the quality of the image by changing the zoom and layer values
        
        19 is a valid zoom value knowing the size of our drone
        
        for the maps you have available
        ROADMAP = "v"
        TERRAIN = "p"
        ALTERED_ROADMAP = "r"
        SATELLITE = "s"
        TERRAIN_ONLY = "t"
        HYBRID = "y"
        
        
    """

    def __init__(self, Northwest, Southeast, zoom=19, layer="s", server="google"):
        """
            MapDownloader Constructor
            Args:
                lat:    The latitude of the location required
                lng:    The longitude of the location required
                zoom:   The zoom level of the location required, ranges from 0 - 23
        """
        if zoom < 0 or zoom > 22:
            raise ValueError("Zoom level must be between 0 and 22")
        self._zoom = zoom
        self._layer = layer

        if server not in ["google", "arcgis"]:
            raise ValueError("Server must be either 'google' or 'arcgis'")
        self._server = server

        # Set up transformers, EPSG:3857 is metric, same as EPSG:900913
        self.to_proxy_transformer = pyproj.Transformer.from_crs('epsg:4326', 'epsg:3857')
        self.to_original_transformer = pyproj.Transformer.from_crs('epsg:3857', 'epsg:4326')
        
        # Check if the coordinates are valid
        if not (isinstance(Northwest, np.ndarray) or isinstance(Northwest, list)) or not (isinstance(Southeast, np.ndarray)  or isinstance(Southeast, list)):
            raise ValueError("Northwest and Southeast must be tuples")
        #make sure northwest and southeast are correctly assigned
        if Northwest[0]<Southeast[0]:
            aux=Northwest[0]
            Northwest[0]=Southeast[0]
            Southeast[0]=aux
        if Northwest[1]>Southeast[1]:
            aux=Northwest[1]
            Northwest[1]=Southeast[1]
            Southeast[1]=aux
        if Northwest[0]<=Southeast[0] or Northwest[1]>=Southeast[1]:
            raise ValueError("you didn't provide a valid rectangle, check your coordinates")
        
        self.Northwest = Northwest
        self.Southeast = Southeast
        
        self.get_tile_coordinates()
        
        if (abs(self.nw_tile[0]-self.se_tile[0]) == 0) or (abs(self.nw_tile[1]-self.se_tile[1]) == 0):
            raise ValueError("Insuficient zoom, points are too close, i am lazy to program this, increase zoom")
        

    def get_tile_coordinates(self):
        """
            Generates an X,Y tile and pixel coordinate based on the latitude, longitude
            and zoom level
        """

        tile_size = 256 #each tile has 256 pixels

        # Use a left shift to get the power of 2, zoom 0 is a world map, zoom 1 is wold map divided in 4 tiles, zoom 2 is world map divided in 16 tiles
        # i.e. a zoom level of 2 will have 2^2 = 4 tiles as coordinate sistem divides axis in 4 (x+y+,x-y+,x-y-,x+y-)
        numTiles = 1 << self._zoom

        ########################
        # For Northwest corner 
        ########################
        # Find the x_pixel given the longitude
        x_pixel = (tile_size / 2 + self.Northwest[1] * tile_size / 360.0) * numTiles 
        # Convert the latitude to radians and take the sine
        sin_y = math.sin(self.Northwest[0] * (math.pi / 180.0))
        # Calulate the y_pixel
        y_pixel = ((tile_size / 2) + 0.5 * math.log((1 + sin_y) / (1 - sin_y)) * -(tile_size / (2 * math.pi))) * numTiles 
        
        self.nw_pixel=[x_pixel, y_pixel]
        self.nw_tile=[int(x_pixel // tile_size), int(y_pixel // tile_size)]
    
        ########################
        # For Southeast corner 
        ########################
        # Find the x_pixel given the longitude
        x_pixel = (tile_size / 2 + self.Southeast[1] * tile_size / 360.0) * numTiles 
        # Convert the latitude to radians and take the sine
        sin_y = math.sin(self.Southeast[0] * (math.pi / 180.0))
        # Calulate the y_pixel
        y_pixel = ((tile_size / 2) + 0.5 * math.log((1 + sin_y) / (1 - sin_y)) * -(tile_size / (2 * math.pi))) * numTiles 
        
        self.se_pixel=[x_pixel, y_pixel]
        self.se_tile=[math.ceil(x_pixel // tile_size)+1, math.ceil(y_pixel // tile_size)+1]
            
    def generateImage(self):
        """
            Generates an image by stitching a number of google map tiles together. after executing self.get_tile_coordinates
            Returns:
                A high-resolution Goole Map image.
        """

        tile_width = abs(self.nw_tile[0]-self.se_tile[0])
        tile_height = abs(self.nw_tile[1]-self.se_tile[1])
        
        # Determine the size of the image, for simplicity, we will print the whole image and later crop it
        width, height = 256 * tile_width, 256 * tile_height

        # Create a new image of the size required
        map_img = Image.new('RGB', (width, height))
        
        
        # calculate northwest pixel of the tile
        start_x=self.nw_tile[0]
        start_y=self.nw_tile[1]

        # print the map and report using tdqm
        pbar = tqdm(total=tile_width*tile_height, unit="tiles")
        pbar.set_description("Downloading map")
        pbar.set_postfix_str(f"Zoom: {self._zoom}, Layer: {self._layer}")
        pbar.refresh()
        for x in range(0, tile_width):
            for y in range(0, tile_height):
                current_tile = str(x) + '-' + str(y)
                try:
                    if self._server == "google":
                        url = f'https://mt0.google.com/vt?lyrs={self._layer}&x=' + str(start_x + x) + '&y=' + str(start_y + y) + '&z=' + str(self._zoom)
                        urllib.request.urlretrieve(url, current_tile)
                    elif self._server == "arcgis":
                        url = f"http://services.arcgisonline.com/ArcGis/rest/services/World_Imagery/MapServer/tile/{self._zoom}/{start_y + y}/{start_x + x}.png"
                        urllib.request.urlretrieve(url, current_tile)
                    else:
                        raise ValueError("Server not available")
                except :
                    raise ValueError("Server not available")
                im = Image.open(current_tile)
                map_img.paste(im, (x * 256, y * 256))

                os.remove(current_tile)
                pbar.update(1)
                pbar.refresh()
        
        print("map downloaded")

        self.raw_map_img = map_img.copy()
        # cut corners
        left=abs(self.nw_tile[0] * 256 - self.nw_pixel[0])
        top=abs(self.nw_tile[1] * 256 - self.nw_pixel[1])
        right=left + abs(self.se_pixel[0] - self.nw_pixel[0])
        bottom=top + abs(self.nw_pixel[1] - self.se_pixel[1])
        
        
        map_img = map_img.crop((left, top, right, bottom))
        self.map_img = map_img

    def image_generator(self):
        """
            Generates an image by stitching a number of google map tiles together. after executing self.get_tile_coordinates
            Returns:
                A high-resolution Goole Map image.
        """

        tile_width = abs(self.nw_tile[0]-self.se_tile[0])
        tile_height = abs(self.nw_tile[1]-self.se_tile[1])
        
        # Determine the size of the image, for simplicity, we will print the whole image and later crop it
        width, height = 256 * tile_width, 256 * tile_height

        # Create a new image of the size required
        map_img = Image.new('RGB', (width, height))
        
        
        # calculate northwest pixel of the tile
        start_x=self.nw_tile[0]
        start_y=self.nw_tile[1]

        # print the map and report using tdqm
        pbar = tqdm(total=tile_width*tile_height, unit="tiles")
        pbar.set_description("Downloading map")
        pbar.set_postfix_str(f"Zoom: {self._zoom}, Layer: {self._layer}")
        pbar.refresh()
        for x in range(0, tile_width):
            for y in range(0, tile_height):
                current_tile = str(x) + '-' + str(y)
                try:
                    if self._server == "google":
                        url = f'https://mt0.google.com/vt?lyrs={self._layer}&x=' + str(start_x + x) + '&y=' + str(start_y + y) + '&z=' + str(self._zoom)
                        urllib.request.urlretrieve(url, current_tile)
                    elif self._server == "arcgis":
                        url = f"http://services.arcgisonline.com/ArcGis/rest/services/World_Imagery/MapServer/tile/{self._zoom}/{start_y + y}/{start_x + x}.png"
                        urllib.request.urlretrieve(url, current_tile)
                    else:
                        raise ValueError("Server not available")
                except :
                    raise ValueError("Server not available")
                im = Image.open(current_tile)
                map_img.paste(im, (x * 256, y * 256))

                os.remove(current_tile)
                pbar.update(1)
                pbar.refresh()
                yield False
        
        print("map downloaded")

        self.raw_map_img = map_img.copy()
        # cut corners
        left=abs(self.nw_tile[0] * 256 - self.nw_pixel[0])
        top=abs(self.nw_tile[1] * 256 - self.nw_pixel[1])
        right=left + abs(self.se_pixel[0] - self.nw_pixel[0])
        bottom=top + abs(self.nw_pixel[1] - self.se_pixel[1])
        
        
        map_img = map_img.crop((left, top, right, bottom))
        self.map_img = map_img
        yield True


    def to_pixel(self, GPS_point):
        """
            Generates an X,Y pixel coordinate based on the latitude, longitude
            and zoom level
            Returns:    An X,Y pixel coordinate
        """
        #check if we are inside the map
        self.pole=[None,None]
        if self.Northwest[0]>self.Southeast[0]:
            self.pole[0]="N"
        else:
            self.pole[0]="S"
        if self.Northwest[1]>self.Southeast[1]:
            self.pole[1]="W"
        else:
            self.pole[1]="E"


        if self.pole[1]=="W":
            if ((self.Southeast[1])>(GPS_point[1])>(self.Northwest[1])):
                raise ValueError(f"point {GPS_point} 1W is outside the map {[self.Northwest,self.Southeast]}")
        else:
            if ((self.Southeast[1])<(GPS_point[1])<(self.Northwest[1])):
                raise ValueError(f"point {GPS_point} 1E is outside the map {[self.Northwest,self.Southeast]}")
        if self.pole[0]=="N":
            if ((self.Southeast[0])>(GPS_point[0])>(self.Northwest[0])):
                raise ValueError(f"point {GPS_point} 0N is outside the map {[self.Northwest,self.Southeast]}")
        else:
            if ((self.Southeast[0])<(GPS_point[0])<(self.Northwest[0])):
                raise ValueError(f"point {GPS_point} 0S is outside the map {[self.Northwest,self.Southeast]}")



        tile_size = 256 #each tile has 256 pixels

        # Use a left shift to get the power of 2, zoom 0 is a world map, zoom 1 is wold map divided in 4 tiles, zoom 2 is world map divided in 16 tiles
        # i.e. a zoom level of 2 will have 2^2 = 4 tiles as coordinate sistem divides axis in 4 (x+y+,x-y+,x-y-,x+y-)
        numTiles = 1 << self._zoom

        # Find the x_pixel given the longitude
        x_pixel = (tile_size / 2 + GPS_point[1] * tile_size / 360.0) * numTiles 

        # Convert the latitude to radians and take the sine
        sin_y = math.sin(GPS_point[0] * (math.pi / 180.0))

        # Calulate the y_pixel
        y_pixel = ((tile_size / 2) + 0.5 * math.log((1 + sin_y) / (1 - sin_y)) * -(
        tile_size / (2 * math.pi))) * numTiles 
        
        return [x_pixel-self.nw_pixel[0], y_pixel-self.nw_pixel[1]]
    
    def show_grid(self, lat_lon_map, base_matrix):

        # if lat_lon_map.shape[0] == base_matrix.shape[0] and lat_lon_map.shape[1] == base_matrix.shape[1]:
        #     print("base_matrix and lat_lon_map have the same shape, extending last column and row")
        #     # Extend the last column and row of lat_lon_map
        #     lat_lon_map = np.concatenate((lat_lon_map, lat_lon_map[-1:, :]), axis=0)
        #     lat_lon_map = np.concatenate((lat_lon_map, lat_lon_map[:, -1:]), axis=1)
        #     for i in range(lat_lon_map.shape[0]):
        #         lat_lon_map[i, -1] = lat_lon_map[i, -2] +( lat_lon_map[i, -2] - lat_lon_map[i, -3])
        #     for i in range(lat_lon_map.shape[1]):
        #         lat_lon_map[-1, i] = lat_lon_map[-2, i] + (lat_lon_map[-2, i] - lat_lon_map[-3, i])
        #     print("lat_lon_map shape: ", lat_lon_map.shape)
            

        # cut corners
        p1=self.to_pixel(lat_lon_map[0, 0])
        p2=self.to_pixel(lat_lon_map[1, 1])
        distance=min(abs(p1[0]-p2[0]),abs(p1[1]-p2[1]))
        left=abs(self.nw_tile[0] * 256 - self.nw_pixel[0] -distance)
        top=abs(self.nw_tile[1] * 256 - self.nw_pixel[1] +distance)
        right=left + abs(self.se_pixel[0] - self.nw_pixel[0] + distance)
        bottom=top + abs(self.nw_pixel[1] - self.se_pixel[1] -distance)
        print("distance: ", distance)


        map_img = self.raw_map_img.crop((left, top, right, bottom))
            
            
        plt.figure()
        plt.imshow(self.map_img, cmap=plt.get_cmap('binary'))
        #obtain xy points of the base matrix
        points = np.argwhere(base_matrix == 1)
                
        for point in points:
            try:
                x=[]
                y=[]
                center = self.to_pixel(lat_lon_map[point[0], point[1]])
                p1=[center[0]-distance/2, center[1]-distance/2]
                p2=[center[0]+distance/2, center[1]-distance/2]
                p3=[center[0]-distance/2, center[1]+distance/2]
                p4=[center[0]+distance/2, center[1]+distance/2]
                x.append(p1[0])
                y.append(p1[1])
                x.append(p2[0])
                y.append(p2[1])
                x.append(p4[0])
                y.append(p4[1])
                x.append(p3[0])
                y.append(p3[1])
                x.append(p1[0])
                y.append(p1[1])
                plt.plot(x,y,'b-', linewidth=distance/20)
            except Exception as e:
                continue
        plt.savefig("map.png", dpi=400)
        plt.show()

    def create_grid_latlon(self, grid_size, name = None):
        """
            Creates a grid of given size over the map
            Args:
                grid_size:  The size of the grid in meters
            Returns:
                A grid of given size over the map.
        """

        # Transform NW and SE points to projection (EPSG:3857)
        self.transformed_nw = self.to_proxy_transformer.transform(self.Northwest[1], self.Northwest[0]) # (x,y)
        self.transformed_se = self.to_proxy_transformer.transform(self.Southeast[1], self.Southeast[0]) # (x,y)
        dx = self.transformed_se[0] - self.transformed_nw[0]
        dy = self.transformed_se[1] - self.transformed_nw[1]
        print(f"map size is {int(abs(dy))}x{int(abs(dx))}m")

        # Calculate the number of pixels per meter at the current zoom level
        meters_per_pixel = 156543.03392 * math.cos(math.radians((self.Northwest[0] + self.Southeast[0]) / 2)) / (2 ** self._zoom)
        pixels_per_grid = grid_size / meters_per_pixel

        grid_img = self.map_img.copy()
        # create an np array with the points (rows, cols) where rows are y (north->south) and cols are x (west->east)
        nrows = int(abs(dy) // grid_size) + 1
        ncols = int(abs(dx) // grid_size) + 1
        grid_latlon = np.zeros((nrows, ncols), dtype=object)
        import PIL.ImageDraw as ImageDraw
        draw = ImageDraw.Draw(grid_img)

        print(f"image shape {grid_latlon.shape}, pixels per grid: {pixels_per_grid}")

        # Determine step directions in projected coordinates (handles any orientation)
        x_dir = 1 if dx >= 0 else -1
        y_dir = 1 if dy >= 0 else -1

        for i in range(nrows):
            for j in range(ncols):
                # Calculate the coordinates in EPSG:3857. Use j as column (x), i as row (y)
                x = self.transformed_nw[0] + j * x_dir * grid_size
                y = self.transformed_nw[1] + i * y_dir * grid_size

                # Transform back to EPSG:4326 (lon, lat)
                lon, lat = self.to_original_transformer.transform(x, y)
                grid_latlon[i, j] = (lat, lon)

                # Draw grid lines in pixel space if requested (note: j -> x, i -> y)
                if i <= nrows - 2 and j <= ncols - 2 and name is not None:
                    # pixel coordinates: (x_pixel, y_pixel)
                    px = j * pixels_per_grid
                    py = i * pixels_per_grid
                    # Draw vertical line from (px,py) to (px, py + pixels_per_grid)
                    draw.line([(px, py), (px, py + pixels_per_grid)], fill=(255, 0, 0), width=1)
                    # Draw horizontal line from (px,py) to (px + pixels_per_grid, py)
                    draw.line([(px, py), (px + pixels_per_grid, py)], fill=(255, 0, 0), width=1)

        if name is not None:
            try:
                np.save(f"{name}_latlon.npy", grid_latlon)
                grid_img.save(f"{name}_grid.png")
                print(f"Grid image saved as {name}_grid.png and latlon grid as {name}_latlon.npy")
            except Exception as e:
                print(f"Error saving files: {e}")
        return grid_latlon
        # Create a new image to draw the grid on

    def save_navigation_waters(self, name="map"):
        """
            Creates a navigation map with water and land
            Args:
                name:  The name of the files to save
            Returns:
                Saves a navigation map with water and land.
        """
        lat_lon = np.load(f"{name}_latlon.npy", allow_pickle=True)
        navigation_map = np.zeros_like(lat_lon, dtype=int)
        occupancy_image = Image.open(f"{name}_grid.png").convert("L")

        for i in range(lat_lon.shape[0]):
            for j in range(lat_lon.shape[1]):
                pixel = occupancy_image.getpixel((j * (occupancy_image.width // lat_lon.shape[1]),
                                                  i * (occupancy_image.height // lat_lon.shape[0])))
                # Assuming water is represented by darker pixels
                if pixel < 128:
                    navigation_map[i, j] = 1  # Water
                else:
                    navigation_map[i, j] = 0  # Land

        np.save(f"{name}_navigation_map.npy", navigation_map)
        print(f"Navigation map saved as {name}_navigation_map.npy")


import tkinter
import tkinter.font
import tkinter.ttk

def create_label(frame, _text, _variable, relief="raised"):
    a= tkinter.Frame(frame, borderwidth=2, relief=relief, padx=10, pady=2)
    a.pack(side='top', fill='both', expand=True,)
    a.pack_propagate(False)
    tkinter.Label(a, text=_text, font=tkinter.font.Font(weight='bold')).pack(side="left", fill='both', padx=0, pady=2)
    tkinter.Label(a, textvariable=_variable).pack(side="left", fill='both', padx=0, pady=2)
    return a
def create_tittle_label(frame, _text, _size=20):
    a= tkinter.Frame(frame,width=400, height=50, borderwidth=2, relief="raised", padx=50, pady=2)
    a.pack_propagate(False)
    tkinter.Label(a, text=_text, font=tkinter.font.Font(weight='bold', size=_size)).pack(side="left", fill='x', padx=0, pady=2)
    return a
    
def create_labelh(frame, _text, _variable, width=10, _size=14, expand="false", fill="both"):
    a= tkinter.Frame(frame, borderwidth=2, relief="raised", padx=10, pady=10)
    a.pack(side='left', anchor="w", fill=fill, expand=expand)
    tkinter.Label(a, text=_text, font=tkinter.font.Font(weight='bold', size=_size)).pack(side="left", padx=0, pady=2,anchor="e")
    ret=tkinter.Label(a , font=tkinter.font.Font(size=_size-2), textvariable=_variable)
    ret.pack(side="left", padx=0, pady=2,anchor="w")
    return ret

def create_label_with_units(frame, _text, _variable, units):
    a= tkinter.Frame(frame, borderwidth=2, relief="raised", padx=10, pady=10)
    a.pack(side='top', fill='both', expand=True,)
    a.pack_propagate(False)
    tkinter.Label(a, text=_text, font=tkinter.font.Font(weight='bold')).pack(side="left", fill='both', padx=0, pady=5)
    tkinter.Label(a, textvariable=_variable).pack(side="left", fill='both', padx=0, pady=5)
    tkinter.Label(a, text=units).pack(side="left", fill='both', padx=0, pady=5)
    return a

def create_labelh_with_units(frame, _text, _variable, units, _size=14, expand="false", fill = "both"):
    a= tkinter.Frame(frame, borderwidth=2, relief="raised", padx=10, pady=10)
    a.pack(side='left', anchor="w", fill=fill, expand=expand)
    tkinter.Label(a, text=_text, font=tkinter.font.Font(weight='bold', size=_size)).pack(side="left", fill=fill, padx=0, pady=5)
    tkinter.Label(a, textvariable=_variable, font=tkinter.font.Font(size=_size-2)).pack(side="left", fill=fill, padx=0, pady=5)
    tkinter.Label(a, text=units, font=tkinter.font.Font(size=_size-2)).pack(side="left", fill=fill, padx=0, pady=5)
    return a




from setuptools import setup

# Minimal shim setup.py — keep pyproject.toml as single source of metadata
# This file exists for legacy tools / `pip install .` and intentionally
# does not duplicate metadata declared in pyproject.toml.

if __name__ == "__main__":
    setup()



import paho.mqtt.client as mqtt
import traceback
import getpass
import threading


class MQTT(object):
    def __init__(self, addr='bender.us.es', port=1883, timeout=5, topics2suscribe=None, on_message=None, on_disconnect=None,user=None,password=None):
        username =getpass.getuser()
        self.client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2, username)

        if topics2suscribe is None:
            topics2suscribe = []
        self.topics2suscribe = topics2suscribe

        if on_message is None:
            self.on_message = self.on_message
        else:
            self.on_message = on_message

        if on_disconnect is None:
            self.on_disconnect = self.on_disconnect
        else:
            self.on_disconnect = on_disconnect
        # Con esto se puede mejorar la seguridad del MQTT si el broker esta configurado para eso
        # self.client.username_pw_set(user, password)
        self._mqtt_thread = threading.Thread(target=self.mqtt_thread, args=(addr, port, timeout,))
        self._mqtt_thread.start()

    def mqtt_thread(self, addr='bender.us.es', port=1883, timeout=5):
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message
        self.client.on_connect = self.on_connect
        self.client.on_disconnect = self.on_disconnect
        try:
            self.client.connect(addr, port, timeout)
        except ConnectionRefusedError:
            print(f"Connection to MQTT server was refused")
            return
        except OSError:
            print(f"MQTT server was not found")
            return
        except TimeoutError:
            print(f"MQTT was busy, timeout error")
            return
        except:
            error = traceback.format_exc()
            print(f"MQTT connection failed, unknown error:\n {error}")
            return
        self.client.loop_forever()
    
    def on_connect(client, userdata, flags, rc):
        print(f"Client {userdata} connected! " + str(rc))

    def on_connect(self,  _client, userdata, flags, rc, properties):
        print("Connected to MQTT server")
        for topic in self.topics2suscribe:
            self.client.subscribe(topic)

    def on_disconnect(self, client, userdata, flags, rc ,properties):
        print("client disconnected ok")

    @staticmethod
    def on_message(_client, user_data, msg):
        message = bool(msg.payload)
        print(message)

    def send_new_msg(self, msg, topic="coordinator"):
        #while not self.client.is_connected():
            #continue
        self.client.publish(topic, msg)

    def close(self):
        self.client.loop_stop()
        self.client.disconnect()


import matplotlib.pyplot as plt
import matplotlib as mpl

fig, ax = plt.subplots(figsize=(1,6), layout='constrained')

colorlist=["#983604", "#CE3202","#E53602","#FE6501","#FDFC0A","#98CC05","#037F03","#3663FE","#0301CE","#300397","#010263","#030153","#020330","#000000"]
cmap = mpl.colors.LinearSegmentedColormap.from_list("dummy",colorlist)
norm = mpl.colors.Normalize(vmin=0, vmax=15)
m=mpl.cm.ScalarMappable(norm=norm, cmap=cmap)

fig.colorbar(m, cax=ax, orientation='vertical', label='Some Units')

print(m.to_rgba(5))

plt.show()


# Programa para pasar de una imagen a una matriz de pixeles 0,1 con un umbral y un alto en pixeles
import sys
from PIL import Image
import numpy as np
import matplotlib.pyplot as plt	
import argparse



def img2pixels(img_path, threshold=128, height=100):

	img = Image.open(img_path)
	img = img.convert('L')  # Convert to grayscale
	img = img.resize((int(img.width * height / img.height), height), Image.LANCZOS)
	img = np.array(img)

	# Threshold
	img = (img > threshold).astype(int)

	return img



if __name__ == "__main__":

	argparser = argparse.ArgumentParser(description='Convert image to pixel matrix')
	argparser.add_argument('--path', type=str, help='Path to image')
	argparser.add_argument('--threshold', type=int, default=128, help='Threshold for binarization')
	argparser.add_argument('--height', type=int, default=100, help='Height of the output image')
	argparser.add_argument('--output', type=str, help='Output file')

	args = argparser.parse_args()

	# Example
	pixels = img2pixels(args.path, threshold=args.threshold, height=args.height)


	# To numpy and save to file 

	np.save(args.output + '.npy', pixels)
